/* Author: Jason Leigh
 * (C) 1997 Electronic Visualization Laboratory - Univ Illinois at Chicago
 */
#include "CAVERN_includes.h++"
#include "CAVERN_irb_c.h++"
#include "CAVERN_irbKey_c.h++"

#ifndef _CAVERN_irbChannel_c
#define _CAVERN_irbChannel_c

/** Class for specifying the characteristics of links.
By default links are auto-synchronizing and use active updating. That is
links will compare local and remote timestamps to determine if
data transfer is needed. Also when data is changed at any site, it is
automatically propagated to remote sites.
  */
class CAVERN_linkAttrib_c {
public:

	/// Choose between ACTIVE or PASSIVE update of linked keys.
	enum CAVERN_update_t {
		/** Active update means key data are synchronized whenever the data changes.
		    This is similar to a server push.
		*/
		ACTIVE,
		/// Passive update means the user must explicitly deliver or request data.
		PASSIVE
	};

	/// Choose between the types of synchronization across the links.
	enum CAVERN_synch_t {
		/// Perform no synchronization at all. Useful for initial synch conditions.
		NONE,

		/// Automatically allow CAVERNsoft to manage synchronization based on content and timestamps.
		AUTO_SYNCH,

		/// Synchronize by tranfering data from local to remote site based on content and time stamp.
		SYNCH_LOCAL_TO_REMOTE,

		/// Synchronize by tranfering data from remote to local site based on content and time stamp.
		SYNCH_REMOTE_TO_LOCAL,

		/// Force tranfer of data from local to remote site regardless of content and timestamp.
		FORCE_LOCAL_TO_REMOTE,

		/// Force tranfer of data from remote to local site regardless of content and timestamp.
		FORCE_REMOTE_TO_LOCAL
	};

	/** This to set how data is initially synchronized between 2 keys when
	    a link is first established.
	    When either SYNCH_LOCAL_TO_REMOTE, SYNCH_REMOTE_TO_LOCAL or AUTO are used local and remote
	    timestamps as well as their contents are checked to decide whether data transfer is necessary.
	    */
	void setInitialSynchType(CAVERN_synch_t t) {
		initialSynchType = t;
	}

	/** This is to set how data is subsequently synchronized between 2 keys in a link.
	    In subsequent synch content nor timestamps are checked. So SYNCH_LOCAL_TO_REMOTE is
	    the same as FORCE_LOCAL_TO_REMOTE. Similarly SYNCH_REMOTE_TO_LOCAL is
	    the same as FORCE_REMOTE_TO_LOCAL. Note. If you use REMOTE_TO_LOCAL
	    synching locally, any changes that you make to your local key will not
	    be propagated to the remote subscriber/client since you have forced data transfer
	    into being unidirectional.
	    These rules can be overriden on temporary bases with: CAVERN_irbLink_c::requestRemote()
	    */
	void setSubsequentSynchType(CAVERN_synch_t t) {
		subsequentSynchType = t;
	}

	/// Returns the current synch type.
	CAVERN_synch_t getSubsequentSynchType() {return subsequentSynchType;}


	/** Set the link's update type to either active or passive.
	    Active update means key data are synchronized whenever the data changes.
	    In some sense this is similar to a server push.
	    Passive update means the user must explicitly deliver or request data.
	*/
	void setUpdateType(CAVERN_update_t t) {
		updateType = t;
	}

	/// Get update type
	CAVERN_update_t getUpdateType() {return updateType;}

	/// Get initial synch type.
	CAVERN_synch_t getInitialSynchType(){return initialSynchType;}

	CAVERN_linkAttrib_c () {
		setInitialSynchType(AUTO_SYNCH);
		setSubsequentSynchType(AUTO_SYNCH);
		setUpdateType(ACTIVE);
	}

private:
	CAVERN_update_t updateType;
	CAVERN_synch_t initialSynchType, subsequentSynchType;


};

/** IRB Communications Channel class. Channels are the actual communication pathways that are
set up between local and remote CAVERNsoft-based clients/servers. These channels
can be set to transmit data via reliable TCP or unreliable UDP. Also at some
point in the future channels can be used
to set the kind of quality of service options (bandwidth, latency, jitter requirements)
desired by your application. Your application can open as many channels as you like,
each possibly with differing communications characteristics. Once these channels are
created you may choose what pieces of data are transmitted over them by linking local
keys with remote keys (see CAVERN_irbLink_c).
  */
class CAVERN_irbChannel_c {
	// So that it can access personalIRB ptr.
	friend CAVERN_irb_c;

public:
	/// Status codes returned from CAVERN_irbChannel_c methods.
	enum status_t  {
		/// Operation succeeded.
		OK,
		/// Operation failed.
		FAILED,
		/// Used for QoS- operation negotiated a lower QoS.
		NEGOTIATED
		// Local Key does not exist.
		//NO_SUCH_LOCAL_KEY,
		// Remote Key does not exist.
		//NO_SUCH_REMOTE_KEY
	};

	CAVERN_irbChannel_c();

	/// Trigger event types for channels.
	enum CAVERN_channelEvent_t  {
		/// Broken link due to break of connection.
		BROKEN_CHANNEL,

		/// Triggers when the QoS of the network service begins to deviate.
		QOS_DEVIATION
	};

	/// Network reliability types
	enum CAVERN_networkReliability_t  {
		/// Use reliable protocols- DEFAULT.
		RELIABLE,
		/// Use unreliable protocols.
		UNRELIABLE
	};

	/// Update types
	enum CAVERN_update_t  {
		/// Active update- DEFAULT.
		ACTIVE_UPDATE,
		/// Passive update.
		PASSIVE_UPDATE
	};

	/** Deleting this channel will close the channel.
	    Also it will unlink all links.
	*/
	~CAVERN_irbChannel_c();

	/** Open the channel to a remote IRB.
	@param CAVERN_irbId_c Specifies remote IRB.

	@param CAVERN_qosAttrib_c Specifies the desired
	QoS. It returns set with the QoS it was able to negotiate.
	Currently this feature has not been implemented since Nexus
	does not yet support QoS capabilities.

	@param CAVERN_networkReliability_t selects either CAVERN_irbChannel_c::RELIABLE or
	CAVERN_irbChannel_c::UNRELIABLE transmission.

	@param status returns OK if got contract ; NEGOTIATED if had to negotiate for lower
	QoS; FAILED if completely failed.
	*/
	void open(CAVERN_irbId_c* remoteIRB, CAVERN_qosAttrib_c* qosAttr,
		  CAVERN_networkReliability_t reliability, CAVERN_irbChannel_c::status_t *retStatus);


	/// Close an open channel.
	void close();

	/** Reserve Network Channel's Quality of Service
	@param CAVERN_qosAttrib_c Specifies the desired network quality of service (QoS).
	It returns set with the QoS it was able to negotiate. If no QoS attributes
	are desired, then set this parameter to NULL.
	@param retStatus returns OK if got contract ; NEGOTIATED if had to negotiate for lower ;
	FAILED if completely failed.
	Currently this feature has not been implemented since Nexus
	does not yet support QoS capabilities.
	*/
	void  reserveQoS(CAVERN_qosAttrib_c *, CAVERN_irbChannel_c::status_t *retStatus);


	/** Link a local and remote key together using this channel.
	    Returns a link object if successful, else NULL.
	    Linking allows you to store data in a local key and have it
	    automatically propagated to a remotely linked key and vice versa.
	    Linking to a non-existant remote key will dynamically create
	    the remote key.
	*/
	CAVERN_irbLink_c * link(CAVERN_irbKey_c* localKey,
				CAVERN_irbKeyId_c* remoteKeyID,
				CAVERN_linkAttrib_c* attr=NULL);

	// Should have an additiona link like:
	// CAVERN_mcastLink_c * linkToMcast(.....);
#if(0)

	// Link a local and remote key together using this channel.
	CAVERN_irbLink_c * link(CAVERN_irbKeyId_c* localKeyID,
			CAVERN_irbKeyId_c* remoteKeyID,
			CAVERN_linkAttrib_c* attr = NULL);
#endif
	/** Trigger on events about the channel.
	  For example, if a connection breaks the user can be warned
	  by a BROKEN_CHANNEL event.

	  @param callback Callback should be of the form:
	  void callback(CAVERN_irbChannel_c::CAVERN_channelEvent_t event,
			 CAVERN_irbChannel_c *thisChannel,
			 void* userData);

	*/
	void trigger(void (*callback)(CAVERN_irbChannel_c::CAVERN_channelEvent_t event,
				      CAVERN_irbChannel_c *thisChannel,
				      void* userData),
		     void* userData);

	void (*triggerCallback)(CAVERN_channelEvent_t event,
			 CAVERN_irbChannel_c *thisChannel,
			 void* userData);

	void *triggerUserData;

	// These should be private but purgeStandardChannelAndItsChannelBuffers
	// in cvrn_comm.c++ needs them.
	enum channelState_t {
		CLOSED,
		OPEN
	};

	// Use this to prevent the user from opening a channel twice.
	channelState_t channelState;

private:


	// This is  how the channel object can remember which bundle & channel it created.
	/*
	cvrnBundle_c *associatedBundle;
	cvrnChannel_c *associatedChannel;

	To get a bundle or a channel you must first look for it in the bundleDB and channelDB.
	this is so that if a channel or bundle gets deleted you don't have a dangling pointer.
	Searching the DBs will reveal null so you can gracefully tell the user that.
	*/
	int associatedBundleId;
	int associatedChannelId;

};
#endif
