/* Author: Jason Leigh
 * (C) 1997 Electronic Visualization Laboratory - Univ Illinois at Chicago
 */

#ifndef _CAVERN_irbKey_c
#define _CAVERN_irbKey_c

#include "CAVERN_includes.h++"
#include "CAVERN_irbKeyId_c.h++"

class cvrnActiveKeyEntry_c;
class CAVERN_irbKeyId_c;


/** CAVERN IRB Key class. Objects of this type are instantiated by
calling CAVERN_irb_c::define(). When you are done refering to the key
delete it to free up system resources.
*/
class CAVERN_irbKey_c {

public:

	/** The destructor will dereference the local key so that it may
	be garbage collected to free up CAVERN's resources. Hence it is a way to
	undefine a key. There is currently no call to undefine a key permanently including
	the persistent store.
	*/
	~CAVERN_irbKey_c();

	/// Status codes returned from CAVERN_irbKey_c methods.
	enum status_t  {
		/// Operation succeeded.
		OK,
		/// Operation failed.
		FAILED,
		/** Access to the key failed because the key is no longer valid;
		    delete the key and redefine it if you wish to access it.
		    */
		KEY_STALE,
		/** Returned if you attempt a get() but did not provide a large
		    enough buffer for CAVERNsoft to store the data into.
		    */
		BUFFER_TOO_SMALL
	};

	/** Used to specify if specific key operations should block till completion
	or not (See put(), deliver(), deliverIncludingPassive()).
	*/
	enum CAVERN_irbKeyBlocking_t {
		/// Blocking.
		BLOCKING,
		/// Non-blocking.
		NON_BLOCKING
	};
	
	// Junna, modified to specify threaded handler or non-threaded handler
	/** Used to specify that when the data eventually reaches its destination,
	    whether to call the key's trigger callback function via a separately
	    spawned thread or not. In general if you have small pieces of data
	    updating rapidly you should use a non-threaded trigger to prevent
	    the IRB from over spawning threads and crashing the IRB.
	    If you intend to infrequently send large pieces of data via a key
	    then a threaded trigger would be desirable.
	    @see put(), import(), deliver(), deliverIncludingPassive().
	*/
	enum CAVERN_irbKeyTrigger_t	{
		/// Non-threaded trigger.
		NON_THREADED_TRIGGER,
		/// Threaded trigger.
		THREADED_TRIGGER
	}; 

	/// Trigger event types for keys.
	enum CAVERN_irbKeyEvent_t  {

		/// New data has arrived at the key.
		NEW_DATA_ARRIVED,

#if(0)
		/* This says that remote data is availabe but is not
		  downloaded.  For passive updates the data is not
		  sent but a notification of a newer version of the
		  data is available is sent.
		  */
		NEW_DATA_REMOTELY_AVAILABLE
#endif

	};


	/** Put data into key. This replaces the previous contents with the
	  new contents. This call usually results in a copy of the entire
	  data to the remote site.
	  @param buffer gets copied to key.
	  @param sze amount of data to copy. Returns sze successfully copied.
	  @param status returns status:
	  FAIL occurs if any general problems occurred (like internal mem allocation)
	  KEY_STALE occurs if this key is no longer valid for data access.
	  OK is returned if the put is successful.
	  @param blockingType if set to BLOCKING will return only after the data has
	  been delivered, otherwise this function will return immediately after having
	  initiated delivery. Blocking is ususually used if you don't wish to
	  continue with the flow of your program until a piece of data has
	  been guaranteed to be delivered to the destination. For example
	  you would perform a blocking put before you did a remoteCommit on a link.
	  @param triggerType if set to THREADED_TRIGGER, the threaded handler will 
	  be called; if set to NON_THREADED_TRIGGER, the non-threaded handler will
	  be called. See CAVERN_irbKeyTrigger_t for more details.
	  */
	// Junna, added triggerType parameter to put()
	void put(char *buffer, int *sze, CAVERN_irbKey_c::status_t *status,
		 CAVERN_irbKey_c::CAVERN_irbKeyBlocking_t blockingType = CAVERN_irbKey_c::NON_BLOCKING,
		 CAVERN_irbKey_c::CAVERN_irbKeyTrigger_t triggerType = CAVERN_irbKey_c::NON_THREADED_TRIGGER);

	/** Put data into key. This replaces the previous contents with the
	  new contents. This call usually results in a copy of the entire
	  data to the remote site.
	  @param buffer gets copied to key.
	  @param sze amount of data to copy. Returns num of bytes successfully copied.
	  @param metaData assigns meta data field of a key.
	    Meta data means data about the data.
	    Meta data is intended
	    to be small. One e.g. of its use is to send type information about the data
	    in a key.
	  @param metaSize size of meta data. Returns num of bytes successfully copied.
	  @param status returns status:
	  FAIL occurs if any general problems occurred (like internal mem allocation)
	  KEY_STALE occurs if this key is no longer valid for data access.
	  OK is returned if the put is successful.
	  @param blockingType if set to BLOCKING will return only after the data has
	  been delivered, otherwise this function will return immediately after having
	  initiated delivery. Blocking is ususually used if you don't wish to
	  continue with the flow of your program until a piece of data has
	  been guaranteed to be delivered to the destination. For example
	  you would perform a blocking put before you did a remoteCommit on a link.
	  @param triggerType if set to THREADED_TRIGGER, the threaded handler will 
	  be called; if set to NON_THREADED_TRIGGER, the non-threaded handler will
	  be called. See CAVERN_irbKeyTrigger_t for more details.
	  */
	// Junna, added triggerType parameter to put()
	void put(char *buffer, int *sze, char* metaData, int* metaSize, CAVERN_irbKey_c::status_t *status,
		 CAVERN_irbKey_c::CAVERN_irbKeyBlocking_t blockingType = CAVERN_irbKey_c::NON_BLOCKING,
		 CAVERN_irbKey_c::CAVERN_irbKeyTrigger_t triggerType = CAVERN_irbKey_c::NON_THREADED_TRIGGER);


	/** Import data into key from a file. This replaces the previous contents with the
	  new contents. This call usually results in a copy of the entire
	  data to the remote site.
	  @param filename specifies the file to load.
	  @param sze this returns the size of the file loaded.
	  @param status returns status:
	  FAIL occurs if any general problems occurred (like internal mem allocation)
	  KEY_STALE occurs if this key is no longer valid for data access.
	  OK is returned if the put is successful.
	  @param blockingType if set to BLOCKING will return only after the data has
	  been delivered, otherwise this function will return immediately after having
	  initiated delivery. Blocking is ususually used if you don't wish to
	  continue with the flow of your program until a piece of data has
	  been guaranteed to be delivered to the destination. For example
	  you would perform a blocking put before you did a remoteCommit on a link.
	  @param triggerType if set to THREADED_TRIGGER, the threaded handler will 
	  be called; if set to NON_THREADED_TRIGGER, the non-threaded handler will
	  be called. See CAVERN_irbKeyTrigger_t for more details.
	  */
	// Junna, added triggerType parameter to import()
	void import(char *filename, int *sze, CAVERN_irbKey_c::status_t *status,
		 CAVERN_irbKey_c::CAVERN_irbKeyBlocking_t blockingType = CAVERN_irbKey_c::NON_BLOCKING,
		 CAVERN_irbKey_c::CAVERN_irbKeyTrigger_t triggerType = CAVERN_irbKey_c::NON_THREADED_TRIGGER);

	/** Import data into key from a file. This replaces the previous contents with the
	  new contents. This call usually results in a copy of the entire
	  data to the remote site.
	  @param filename specifies the file to load.
	  @param sze this returns the size of the file loaded.
	  @param metaData assigns meta data field of a key.
	    Meta data means data about the data.
	    Meta data is intended
	    to be small. One e.g. of its use is to send type information about the data
	    in a key.
	  @param metaSize size of meta data. Returns num of bytes successfully copied.
	  @param status returns status:
	  FAIL occurs if any general problems occurred (like internal mem allocation)
	  KEY_STALE occurs if this key is no longer valid for data access.
	  OK is returned if the put is successful.
	  @param blockingType if set to BLOCKING will return only after the data has
	  been delivered, otherwise this function will return immediately after having
	  initiated delivery. Blocking is ususually used if you don't wish to
	  continue with the flow of your program until a piece of data has
	  been guaranteed to be delivered to the destination. For example
	  you would perform a blocking put before you did a remoteCommit on a link.
	  @param triggerType if set to THREADED_TRIGGER, the threaded handler will 
	  be called; if set to NON_THREADED_TRIGGER, the non-threaded handler will
	  be called. See CAVERN_irbKeyTrigger_t for more details.
	  */
	void import(char *filename, int *sze, char* metaData, int *metaSize,
		    CAVERN_irbKey_c::status_t *status, 
		    CAVERN_irbKey_c::CAVERN_irbKeyBlocking_t blockingType = CAVERN_irbKey_c::NON_BLOCKING,
		    CAVERN_irbKey_c::CAVERN_irbKeyTrigger_t triggerType = CAVERN_irbKey_c::NON_THREADED_TRIGGER);

	/** Export a key to an output file.
	    @param status returns either OK or FAILED.
	    KEY_STALE is returned if this key is no longer valid for data access.

	    @param filename is output file to export to.
	    */
	void export(char *filename, CAVERN_irbKey_c::status_t *status);

	/* Get data from key.
	  @param sze passed as pointer. Returns sze of data returned.
	  If key has no data yet, sze returns 0.
	  @param blockingAttrb determines if the get operation will block
	  if no data is in the key. Default is blocking.
	  @return data buffer of key. It is the user's responsibility to
	  deallocate the memory when done. Returns NULL if key has no data yet.

	  char* get(int *sze, CAVERN_blocking_t blockingAttrib = BLOCKING);
	  */

	/** Get a copy of the data from key.
	  @param bufferToFill is a user provided buffer into which the data is copied.
	  @param sze is the size of the data buffer provided. Sze returns the
	  size of the data copied into the buffer.
	  @param status returns status:
	  BUFFER_TOO_SMALL occurs when
	  the data in the key is larger than data buffer provided by user.
	  OK if everything was fine.
	  KEY_STALE if for some reason this key is no longer active.
	  */
	void get(char *bufferToFill, int *sze, CAVERN_irbKey_c::status_t *status);

	/** Get the size of the meta data.
	    Meta data means data about the data.
	    Meta data is intended
	    to be small. One e.g. of its use is to send type information about the data
	    in a key.
	    */
	void getMetaSize(int *sze, CAVERN_irbKey_c::status_t *status);

	/** Set the meta data field of a key.
	    @param meta assigns meta data field of a key.
	    Meta data means data about the data.
	    Meta data is intended
	    to be small. One e.g. of its use is to send type information about the data
	    in a key.
	    @param size is size of meta data. Returns num of bytes successfully copied.
	*/
	void setMeta(char *meta, int *size, CAVERN_irbKey_c::status_t *status);

	/** Get a copy of the meta data from key.
	    Meta data means data about the data.
	    Meta data is intended
	    to be small. One e.g. of its use is to send type information about the data
	    in a key.
	  @param bufferToFill is a user provided buffer into which the meta data is copied.
	  @param sze is the size of the data buffer provided. Sze returns the
	  size of the meta data copied into the buffer.
	  @param status returns status:
	  BUFFER_TOO_SMALL occurs when
	  the meta data in the key is larger than data buffer provided by user.
	  OK if everything was fine.
	  KEY_STALE if for some reason this key is no longer active.
	  */
	void getMeta(char *bufferToFill, int *sze, CAVERN_irbKey_c::status_t *status);


	/// Get size of data in key.
	void getSize(int *sze, CAVERN_irbKey_c::status_t *status);

	/** Get a copy of the data in the key but auto allocate memory for the buffer.
	    I.e. you give it a plain pointer and it will return a chunk of memory.
	    The returned buffer must be manually deleted by you. If memory
	    could not be allocated the buffer returns NULL, the size returns 0.
	    @param status returns OK if everything was fine. Returns FAILED if
	    a memory allocation problem occurred.
	    Returns KEY_STALE if the this key is no longer valid.
	*/
	void getAutoAlloc(char *&autoAllocatedBuffer, int *sze, CAVERN_irbKey_c::status_t *status);

	/** Get the direct pointer to the data in the key.
	    Use this with CAUTION. This forces a lock on CAVERNsoft's internal
	    database so it could potentially prevent new incoming data from arriving.
	    Use the data and release the resource with releaseDirect() as quickly as
	    possible.
	    @see releaseDirect()
	    @param status returns OK or KEY_STALE.
	*/
	void getDirect(char *&direct, int *sze, CAVERN_irbKey_c::status_t *status);

	/** Similar to getDirect. This function allocates memory in the key
	    of a size specified by the sze parameter. Sze also returns the
	    sze of the data successfully allocated.
	    Direct is a pointer to the allocated memory.
	    Use this with CAUTION. This forces a lock on CAVERNsoft's internal
	    database so it could potentially prevent new incoming data from arriving.
	    Use the data and release the resource with releaseDirect() as quickly as
	    possible.
	  
	    Typical use of this function is:

	    aKey->allocDirect(buffer,....);

	    fillBufferWithData(buffer);

	    aKey->releaseDirect();

	    aKey->deliver(......);

	    Warning. There is currently nothing that guarantees that
	    the data in the key won't be updated by new incoming data
	    between the time of releaseDirect() and deliver().

	    @see releaseDirect()
	    @param status returns OK, KEY_STALE or FAILED
	*/
	void allocDirect(char *&direct, int *sze, CAVERN_irbKey_c::status_t *status);

	/** Release the pointer to the data in the key. See getDirect().
	    At the present time, multiple successive calls to releaseDirect
	    can produce undefined behavior.
	    @param status returns either OK or KEY_STALE.
	    */
	void releaseDirect(CAVERN_irbKey_c::status_t *status);

	/** Commit the local key to persistent store.
	    If you wish to commit a remote key to persistent store see CAVERN_irbLink_c::commitRemote().
	    */
	void commit(CAVERN_irbKey_c::status_t *status);

	/** Dispatch data inside current key.
	    It ensures that if this key is participating in a link it will dispatch
	    its changes to all the remote connections.
	    This is good to do after data is modified with getDirect and releaseDirect since
	    these calls have no knowledge of whether you have modified the key or not.

	    Delivery abides by the constraints set by CAVERN_linkAttrib_c
	    during the creation of the link.

	    See CAVERN_irbLink_c and CAVERN_linkAttrib_c.
	    @param status returns either OK or KEY_STALE.

	  @param blockingType if set to BLOCKING will return only after the data has
	  been delivered, otherwise this function will return immediately after having
	  initiated delivery. Blocking is ususually used if you don't wish to
	  continue with the flow of your program until a piece of data has
	  been guaranteed to be delivered to the destination. For example
	  you would perform a blocking put before you did a remoteCommit on a link.
	  @param triggerType if set to THREADED_TRIGGER, the threaded handler will 
	  be called; if set to NON_THREADED_TRIGGER, the non-threaded handler will
	  be called.  See CAVERN_irbKeyTrigger_t for more details.
	  */
	// Junna, added triggerType parameter to deliver()
	void deliver(CAVERN_irbKey_c::status_t *status,
		     CAVERN_irbKey_c::CAVERN_irbKeyBlocking_t blockingType = CAVERN_irbKey_c::NON_BLOCKING,
			 CAVERN_irbKey_c::CAVERN_irbKeyTrigger_t triggerType = CAVERN_irbKey_c::NON_THREADED_TRIGGER);

	/** This does the same thing as deliver except it delivers to passive links
	as well. See CAVERN_irbLink_c.
	@param status returns either OK or KEY_STALE.

	  @param blockingType if set to BLOCKING will return only after the data has
	  been delivered, otherwise this function will return immediately after having
	  initiated delivery. Blocking is ususually used if you don't wish to
	  continue with the flow of your program until a piece of data has
	  been guaranteed to be delivered to the destination. For example
	  you would perform a blocking put before you did a remoteCommit on a link.
	  @param triggerType if set to THREADED_TRIGGER, the threaded handler will 
	  be called; if set to NON_THREADED_TRIGGER, the non-threaded handler will
	  be called. See CAVERN_irbKeyTrigger_t for more details.
	*/
	// Junna, added triggerType parameter to deliverIncludingPassive()
	void deliverIncludingPassive(CAVERN_irbKey_c::status_t *status,
				     CAVERN_irbKey_c::CAVERN_irbKeyBlocking_t blockingType = CAVERN_irbKey_c::NON_BLOCKING,
					 CAVERN_irbKey_c::CAVERN_irbKeyTrigger_t triggerType = CAVERN_irbKey_c::NON_THREADED_TRIGGER);

	// This is to be private and should not be used by the user. It actually does the work of the delivery.
	// Junna, added triggerType parameter to doDeliver()
	void doDeliver(CAVERN_irbKey_c::status_t *status, int checkActiveParam, int
waitForAcknowledge, CAVERN_irbKey_c::CAVERN_irbKeyTrigger_t triggerType);

	/** Trigger on an event.
	    This will call a user-provided callback when an event occurs. See CAVERN_irbKeyEvent_t.
	    WARNING: Currently you cannot perform a put() or any of the deliver() calls on this
	    same key while you are in
	    the user's callback [get() calls are ok.] This is because the key is locked to protect other clients
	    from writing to your key while you are reading it. Hopefully soon we will
	    come up with a solution for this.

	    Also note that when this CAVERN_irbKey_c object is deleted the
	    trigger is also removed. Even if you have other CAVERN_irbKey_c objects
	    that are defined to the same path and key, the trigger is always
	    associated with the object that performed the trigger() call.
	  @param callback is the callback to call when trigger occurs.
	  Callback is of the form:
	  void myCallback(CAVERN_irbKey_c::CAVERN_irbKeyEvent_t event, CAVERN_irbKey_c *thisKey,
						void* userData);
	  @param userData is address of data to pass into callback.
	  */
	void trigger(void (*callback)(CAVERN_irbKey_c::CAVERN_irbKeyEvent_t event,
				      CAVERN_irbKey_c *thisKey,
				      void* userData),
		     void* userData);

	/// Get a copy of the key Id that defined this key.
	CAVERN_irbKeyId_c getKeyId();

	CAVERN_irbKey_c();

	// Strictly speaking this is private.
	// Use this active key to lookup the active key entry in the active key manager.
	// No raw pointers allowed: cvrnActiveKeyEntry_c *activeKey;
	md5Key_c activeKey;

	// Keep a copy of the keyid that created this key.
	CAVERN_irbKeyId_c keyId;

private:

	// Undefine a key. This is only called by the destructor.
	void undefine();

	friend CAVERN_irb_c;

};


#if(0)

/* These are member functions that have not been implemented */

	/* Suspend active updates. You can hence use this as a form of flow control
	for incoming data.
	@see resumeActiveUpdates
	*/
	void suspendActiveUpdates();

	/* Resume active updates.
	  @see suspendActiveUpdates
	  */
	void resumeActiveUpdates();


	/* Get subpart of data from key.
	  Note, this usually implies the data is rather large so some form of
	  a "sparse virtual memory emulation" will be used to "page-in" segments of
	  the imported data.
	  @param startIndex beginning of subsection to extract.
	  @param endIndex end of subsection to extract. If endIndex is NULL it means
	  extract to the end. endIndex must be either NULL or greater than or equal to
	  startIndex.
	  @param sze passed as pointer. Returns sze of data extracted.
	  If key has no data yet returns 0.
	  @return data buffer of subsection. It is the user's responsibility to
	  deallocate the memory when done.
	  */
	char* get(int startIndex, int endIndex, int *sze,
		  CAVERN_blocking_t blockingAttrib);

	// Implemented in the CAVERN_link_c class instead.
	/* Request the data from a remote source over a specific link.
	    Used for passive updates. This call will be ignored for active updates.
	 */
	//void request(CAVERN_link_c *aLink, CAVERN_blocking_t blockingAttrib);

	// Request subpart of data from remote source over a specific link.
	//void request(CAVERN_link_c *aLink, int startIndex, int endIndex,
	//     CAVERN_blocking_t blockingAttrib);



	/* Put data into subpart of key. This usually results in the
	  transmission of the subpart rather than the whole piece of data since
	  this is likely to be used to access a large data set.
	  If Put is done when there is incoming data then the put
	  places the data in the local key, and ignores the incoming
	  data stream if the local timestamp is more current.
	  @param buffer gets copied to key.
	  @param startIndex is location in key to begin copying.
	  @param sze amount of data to copy. Returns sze successfully copied.
	  If sze exceeds current sze of data in key, data cell is automatically expanded.
	  */
	void put(char *buffer, int startIndex, int *sze);

	/* Append data to data in key. Only the newly appended extra data
	  is transmitted. Note when you use calls like Append and Put
	  you run the risk of inconsistencies when there are multiple clients
	  updating the same information. It is advisable to use locks in those
	  situations.
	  @param buffer copied and appended to data in key.
	  @param sze amount of data to append. Returns sze successfully appended.
	  */
	void append(char *buffer, int *sze);


	/* Get info about this key.
	  @param keyInfo will return information about the state of the key.
	  @see CAVERN_keyInfo_c for an explanation of status.
	  */
	CAVERN_irbKeyInfo_c getInfo();

	// Look at size of data in key.  getInfo will do this too but with more detail.
	int getSize();

	/* Trigger on an event.
	  @param callback is the callback to call when trigger occurs.
	  Callback is of the form:
	  void (*callback)(CAVERN_triggerEvent_t event, CAVERN_key_c *thisKey,
						void* userData);
	  @param userData is address of data to pass into callback.
	  */
	void trigger(void (*callback)(CAVERN_keyTriggerEvent_t event,
				      CAVERN_irbKey_c *thisKey,
				      void* userData),
		     void* userData);

	/* Lock this key. 
	  The lock works as follows: First this call blocks while it checks the lock.
	  If it is not locked then it grabs the lock.
	  If the item is already locked it
	  can be set to block further until lock becomes available or return
	  with a failure to lock.

	  Note: Eventually we will also provide read/write locks to improve
	  concurrency.

	  @return The return value is status of the lock attempt: either OK or FAILED
	  */
	CAVERN_irbKey_c::status_t lock(CAVERN_blocking_t blockingParam);


	/* Unlock a previously acquired lock.
	  @see lock().
	  */
	void unlock();

	/* Tell IRB to manage the queue using data from this key.
	  WARNING: For large keys this can eat up virtual memory fast!
	  Deleting the queue means you no longer want the IRB to manage
	  it.
	  */

	CAVERN_discreteQueue_c * queue(CAVERN_irbKey_c::status_t status);

	/* Tell IRB to manage stream using data from this key.
	  WARNING: For large keys this can eat up virtual memory fast!
	  Deleting the stream means you no longer want the IRB to manage it.
	 */
	CAVERN_stream_c *queue(CAVERN_irbKey_c::status_t status);
#endif

#endif
