/* Author: Jason Leigh
 * (C) 1997 Electronic Visualization Laboratory - Univ Illinois at Chicago
 */

#ifndef _CAVERN_IRBLINK_C
#define _CAVERN_IRBLINK_C

#include "cvrn_base.h++"
#include "generic_package.h++"
#include "CAVERN_irbChannel_c.h++"

/** Link management class.
This class allows you to manage a link. It allows you to commit the data in
a remote key to a persistent data-store. It allows you to explicitly request data
from a remote key rather than having to wait for data to be automatically
propagated to you. More features will be added with time.
*/
class CAVERN_irbLink_c {

public:

	/// Status codes returned from CAVERN_irbLink_c member functions.
	enum status_t {
		/// Operation failed
		FAILED,
		/// Operation succeeded
		OK,
		/// Operation timed out
		TIMED_OUT,
		/// This is a status message returned from a remote request if the synchronizing time stamps determine that a remote download was not necessary.
		NO_DOWNLOAD_NEEDED
	};

	/// Used to set either blocking or non-blocking remote request.
	enum CAVERN_linkBlocking_t {
		/// Non blocking remote request.
		NON_BLOCKING,
		/// Blocking remote request.
		BLOCKING
	};

	/// Trigger event types for links.
	enum CAVERN_linkEvent_t {

		/// This event occurs when there is news about a remote request.
		REMOTE_REQUEST
	};


	/** Set a user-specified callback to be called when an event occurs.
	    Callback should be of the form:

	    void callback(CAVERN_irbLink_c::CAVERN_linkEvent_t event,
	    CAVERN_irbLink_c *thisLink,
	    CAVERN_irbLink_c::status_t status,
	    void* userData);

	    This callback may receive for e.g. a REMOTE_REQUEST event with a status of
	    TIMED_OUT.

	    @param userData is for passing any user data to the callback on invocation.


	    */
	void trigger(void (*callback)(CAVERN_irbLink_c::CAVERN_linkEvent_t event,
				      CAVERN_irbLink_c *thisLink,
				      CAVERN_irbLink_c::status_t status,
				      void* userData),
		     void* userData);
		     


	//Lock the remote key that is linked to this key (if any).
	//status_t lockRemote(CAVERN_blocking_t blockingParam);

	// UnLock the remote key that is linked to this key.
	//status_t unlockRemote();

	/// This destructor will also unlink the link.
	~CAVERN_irbLink_c();

	/** Tell the remote IRB to commit the remote key's data to persistent data store.
	    @param status returns either FAILED or OK.
	    */
	void commitRemote(status_t *status);

	/// Remote request types.
	enum CAVERN_requestType_t {
		/// Default request will compare timestamps.
		DEFAULT,
		/// Forced request will always download data regardless of timestamps.
		FORCED_REQUEST
	};

	/** Request the data from a remote source over a specific link.
	    Used primarily for passive updates.
	    Callback of remote request could get status:
	    FAILED, NO_DOWNLOAD_NEEDED, TIMED_OUT or OK.
	    OK event means data will be transmitted to your locally linked
	    key. This will fire the irb Key's callback for new incoming data.
	    @param timeOutTime is the time (in seconds) this request waits for either
	    status information or incoming data. When timeout occurs a TIME_OUT
	    event will fire.
	    @param blocking is used to make this request a blocking or non-blocking request.
	    If non-blocking then this function will return immediately with either
	    a status of OK or FAILED that reflects the status of calling
	    this request (ie not the status of the result of the completed
	    request).
	    The callback will actually reveal the true status of the completed request.
	    If blocking then this function waits till remote status information arrives
	    and remote data is completely downloaded (whichever occurs first). The
	    status in this case would be the status of the completed request- i.e.
	    the status the callback would get.
	 */
	void  requestRemote(CAVERN_requestType_t requestType, status_t *status,
			    CAVERN_irbLink_c::CAVERN_linkBlocking_t blocking = CAVERN_irbLink_c::NON_BLOCKING,
			    int timeOutTime = 30);

	// Should be private. It will be cleanedup in the future.
	void (*userCallback)(CAVERN_linkEvent_t event,
			   CAVERN_irbLink_c *thisLink,
			   CAVERN_irbLink_c::status_t status,
			   void *userData);
	void *userData;

	// Mutex to make sure all link calls for a link are sequential and atomic.
	nexus_mutex_t linkOperationMutex;

private:

	int active;

	friend CAVERN_irbChannel_c;

	CAVERN_irbLink_c(md5Key_c key, int bundId, int chanId);

	// Id  of this link.
	md5Key_c channelBufferId;

	// Id of the bundle and channel that this link is associated.
	int channelId;
	int bundleId;

	// Thread to do timeout signalling of remoteRequest calls.
	nexus_thread_t timeoutThread;

	genericPackage paramPack;


	// Unlink the local and remote keys.
	void unLink();



};


#endif
