#ifndef _CAVERNPLUS_CONDITION_C
#define _CAVERNPLUS_CONDITION_C

#include "CAVERN.h++"

/**
Class for thread condition variables.
Condition variables are used in conjunction with
mutexes to provide a way for threads to wait until
a condition occurs.

An example of waiting on a signal is:

\begin{verbatim}

// Lock your mutex that is protecting someState.
myMutex->lock();

// Watch for your desired state to occur.
while(someState != reached) {

	// Wait for a signal.
	myCondition->wait(myMutex);

	.... got the condition and the lock so now continue ....

}

myMutex->unlock();

\end{verbatim}

An example of sending the signal is:

\begin{verbatim}

// Lock your mutex that is protecting someState.
myMutex->lock();

// Signal that the state has been reached.
if (someState == reached) myCondition->signal();

// Unlock your mutex so that the waiting thread can continue.
myMutex->unlock();

\end{verbatim}
*/

class CAVERNplus_condition_c {
public:

	/** Wait on a condition to be signalled.
	    This function first releases the mutex and then
	    waits on the condition. When the condition
	    arises (ie it has been signaled) the mutex is
	    reaquired, and the function returns.
	    @return 0 if function successfully completes else non-zero
	    */
	int wait(CAVERNplus_mutex_c *mutex);

	/** Signal that a condition has arisen. This wakes up
	    one thread that is suspended on this condition. If
	    no threads are suspended this call has no effect.

	    @return 0 if function successfully completes else non-zero


	    */
	int signal();

	/** Signal that a condition has arisen. This wakes up
	    ALL threads that are suspended on this condition. If
	    no threads are suspended this call has no effect.

	    @return 0 if function successfully completes else non-zero
	    */
	int broadcastSignal();

	/// Return nexus' condition variable.
	nexus_cond_t * getNexusCondition();

	CAVERNplus_condition_c();
	~CAVERNplus_condition_c();

private:
	nexus_cond_t itsCV;

};


#endif

