
#ifndef _CAVERNPLUS_DATAPACK_C_H
#define _CAVERNPLUS_DATAPACK_C_H

#include <stdio.h>
#include <string.h>


/**
Data packing class. It is basically a glorified memcpy().
The idea is that you create an object
to help you pack data for transmission over networks.

First you create a CAVERNplus_datapack_c object. Then assign it a pre-allocated
memory buffer. Then using the various pack*() member functions, you can pack
integers, characters, floats etc into the buffer. When done you can simply
assign the buffer to whatever call needs the buffer.

Similarly if you receive a buffer of data from the network, you can 
unpack its constituent components using the unpack*() member functions.
Note: this class does not save the format of your packing or unpacking it simply
lays out your data in the buffer you provide it.
You must pack things and unpack things in the same order in order for them
to make sense.

@author: Jason Leigh

@version: 3/7/97


*/
class CAVERNplus_datapack_c {


public:
	CAVERNplus_datapack_c() {
		buffer = 0; size = 0;
		lastOperation = 1;
	}

	/// User must specify a pre-allocated data buffer and size of the buffer.
	CAVERNplus_datapack_c(char *buff, int sz){
		buffer = buff;
		size = sz;
		lastOperation = 1;
		running = buffer;
	}

	/// User can switch data buffers. But again you need to specify the size.
	void setBuffer(char *buff, int sz) {
		buffer = buff; size = sz;
		running = buffer;
		lastOperation = 1;
	}

	/// Retrieve the data buffer.
	char *getBuffer() { return buffer;}

	/// Retrieve the original full size of the data buffer.
	int getBufferMaxSize() { return size;}

	/// Retrieve the size of the data buffer that has been packed with data so far.
	int getBufferFilledSize() {return running - buffer;}

	/// Reset the packing position to the beginning of the buffer.
	void restart() {lastOperation = 1; running = buffer;}

	/// Floats
	int packFloat(float val) { return pack((char *)&val,sizeof(float));}

	/// Long
	int packLong(long val) {return pack((char *)&val,sizeof(long));}

	/// Int
	int packInt(int val) {return pack((char *)&val,sizeof(int));}

	/// Double
	int packDouble(double val) {return pack((char *) &val,sizeof(double));}

	/// Char
	int packChar(char val) {return pack((char *)&val,sizeof(char));}

	/// Raw char*
	int pack(char *buf, int sz) {

		// If the previous operation was a pack then reset
		// packing pointer to start of buffer.
		if (lastOperation == 1) {
			running = buffer;
		}

		if (running >= (buffer+size)) {
			printf("CAVERNplus_datapack_c: data packing exceeded buffer space.\n");
			return -1;
		} else {

			// Pack the data.
			lastOperation = 0;
			memcpy(running,buf,sz);
			running+= sz;
			return sz;
		}
	}


	/// Float
	float unpackFloat() {float val; unpack((char*) &val, sizeof(float)); return val;}

	/// Long
	long unpackLong() {long val; unpack((char*) &val, sizeof(long)); return val;}

	/// Int
	int unpackInt() {int val; unpack((char*) &val, sizeof(int)); return val;}


	/// Double
	double unpackDouble()  {double val; unpack((char*) &val, sizeof(double)); return val;}

	/// Char
	char unpackChar() {char val; unpack((char*) &val, sizeof(char)); return val;}

	/// Char*
	int unpack(char *buf, int sz) {

		// If last operation was Pack, then reset packing pointer to
		// the start.
		if (lastOperation == 0) {
			running = buffer;
		}

		if (running >= (buffer+size)) {
			printf("CAVERNplus_datapack_c: reading from data packing buffer exceeding buffer space.\n");
			return -1;
		}else {

			// Unpack data.
			lastOperation = 1;
			memcpy(buf, running, sz);
			running+= sz;
			return sz;
		}
	}

private:
	char *buffer;
	int size;
	char *running;
	int lastOperation;
	// 1 = unpack
	// 0 = pack

	
};
#endif
