/* Author: Jason Leigh
 * (C) 1997 Electronic Visualization Laboratory - Univ Illinois at Chicago
 */
#include "CAVERN_includes.h++"
#include "cvrn_comm.h++"
#include "cvrn_db.h++"

#ifndef _CAVERN_irb_c
#define _CAVERN_irb_c

#include "CAVERNInit.h++"
#include "CAVERN_irbId_c.h++"

/** Class for the personal IRB returned from CAVERNInit().
CAVERNInit() is the only function that generates this object.
Each client or server program only has 1 of these objects.

Once you have this IRB object you can call the CAVERN_irb_c::define()
member function to define local keys in which data may be stored.

Then you can create a channel (CAVERN_irbChannel_c) between a local
and remote IRB and link (CAVERN_link_c) keys over this channel so that
any updates to a key will be shared across the network.

*/
class CAVERN_irb_c {

private:
	friend CAVERN_irb_c * CAVERNInit(int *argc, char*** argv, CAVERN_initAttrib_c *initAttr);

	CAVERN_irbId_c selfIrbId;

public:

	/// Find out what your IRB ID is (in terms of initial IP and port.)
	CAVERN_irbId_c getSelfIRBId();

	/// Status codes returned from various methods.
	enum status_t  {
		/// Operation successful.
		OK = -1,

		// Key does not exist.
		//KEY_NOT_EXIST = -2,

		// Cannot connect to remote IRB.
		//CANNOT_CONNECT_TO_REMOTE_IRB = -3,

		// Permission denied  during subscription.
		//PERMISSION_DENIED = -4,

		// Key exists.
		//KEY_EXISTS = -5,

		/// Operation failed.
		FAILED = -6,

		/// Key does not exist anywhere- not in persistent store, nor in the current database.
		KEY_NOT_DEFINED = -7,

		/// Key exists in persistent store but has not been loaded into current database.
		KEY_PERSISTENT_NOT_DEFINED = -8,

		/// Key is persistent and is currently in the database.
		KEY_PERSISTENT_DEFINED = -9,

		/// Key is not persistent but is currently in the database.
		KEY_TRANSIENT_DEFINED = -10
	};

#if(0)
	/* Trigger events for IRBs.
	  */
	enum CAVERN_irbTriggerEvent_t  {
		// Event for when a connection is broken.
		CONNECTION_BROKEN,

		// Remote IRB has cleanly shutdown.
		SHUTDOWN
	};
#endif

	/** Deleting the IRB will shut it down. It will also
	    shutdown your application- Nexus won't let me
	    keep the application alive.
	*/
	~CAVERN_irb_c();

	/** Create a channel (either TCP or UDP)
	    over which local and remote keys may share data.
	    After a channel is created it can be opened to connect to a remote
	    IRB. See CAVERN_irbChannel_c::open().
	    @return a CAVERN_irbChannel_c object or NULL if failed.
	  */
	CAVERN_irbChannel_c *createChannel();

	/** Create a multicast channel to a multicast group.
	    After a channel is created it can be opened to connect to the mcast group
	    by using CAVERN_irbMcastChannel_c::open().
	    @return a CAVERN_irbMcastChannel_c object or NULL if failed.
	  */
	CAVERN_irbMcastChannel_c *createMcastChannel();

	/** Define a key on the personal IRB.

	  @param keyDef if set to NULL then it assumes default settings.
	  Currently this is not used so please set it to NULL.

	  @return A handle to the database. Delete the handle when you're
	  done referencing it. It will help free up system resources.

	  @param status returns either: OK or FAILED.
	  */
	CAVERN_irbKey_c *define(CAVERN_irbKeyId_c *newKeyInfo,
				CAVERN_keyDefAttrib_c *keyDef,
				CAVERN_irb_c::status_t *status);


	/** Determine whether a key has previously been defined.
	    Keys can either be transient or persistent as well as defined or not.
	    Persistent keys are keys that have an actual record in the persistent
	    store (ie disk). Transient key are keys that in the current database cache
	    but has never been committed to persistent store.

	    @param newKeyInfo is the key you want to find out about.
	    @param status returns either: KEY_NOT_DEFINED, KEY_PERSISTENT_DEFINED,
	    KEY_TRANSIENT_DEFINED, KEY_PERSISTENT_NOT_DEFINED.
	  */

	void existsKey(CAVERN_irbKeyId_c *newKeyInfo, CAVERN_irb_c::status_t *status);

	// Perform the actual work of defining the key. This should really
	// be considered private. It is used by cvrn_comm and should not
	// be used by the user.
	void defineAux(CAVERN_irbKeyId_c *newKeyInfo,
				CAVERN_keyDefAttrib_c *keyDef,
				status_t *status);


#ifdef _CVRN_ALLOW_REMOTE_DEFINE
	/* Define a key on a remote IRB.  By defining a remote key
	  that key is declared PERSISTENT.  That means it will
	  continue to exist even when the IRB is stopped and
	  restarted. The only way to remove the key is to undefine it.
	  This simplifies issues on crash recovery.

	 This function does not return a key as you might expect
	 because you must link (see CAVERN_irbChannel_c & CAVERN_link_c) to it
	 before you can access data from it. To do this you need to define a
	 local key so that the remote key and the local can be linked.
	 To define a local key see the other overloading of this define() function.

	 If you define a key that already exists and the attributes
	 of your new definition are different then you will change
	 those attributes on the remote site providing you have the permissions.

	 Defining a key that already exists does NOT clear out its cached contents.

	 @param keyDef set this to NULL for default settings.
	 This is intended to set characteristics such as permissions and
	 frequency and conditions of persistent update.

	 @param status Returns one of the following status codes: OK,
	 CANNOT_CONNECT_TO_REMOTE_IRB, EXISTS, PERMISSION_DENIED
	 */
	void define(CAVERN_irbId_c *remoteIRB,
		    CAVERN_irbKeyId_c *newKeyInfo,
		    CAVERN_keyDefAttrib_c *keyDef,
		    status_t *status);
#endif


};



#endif
