/* Author: Jason Leigh
 * (C) 1997 Electronic Visualization Laboratory - Univ Illinois at Chicago
 */

#ifndef _CVRN_COMM
#define _CVRN_COMM

#include "cvrn_nexus.h"
#include "cvrn_base.h++"
#include "_support_hashDict.h++"
#include "CAVERN_includes.h++"
#include "CAVERN_irbChannel_c.h++"
#include "CAVERN_irbLink_c.h++"


#define _CVRN_DEFAULT_BUNDLE_CHANNEL_CREATION_TIMEOUT_TIME 60

/* This is how much time is allowed to pass for a link to be created.
If this is too small and you are linking 2 keys with very big files attached
to them then it could time out. Originally I had this set to 5 which failed for
10Meg files that are NFS mounted. So now I'm setting it to 60.
*/
#define _CVRN_DEFAULT_MAKE_LINK_TIMEOUT_TIME 60

///const unsigned short _cvrnDefaultHailingChannelPort = 10000; NOT USED.
// Even though these are maxes they are assigned to the managedArray template class
// which has the ability to autogrow when the array fills. These maxes should therefore
// be considered initial defaults.

//----------------------------------------------------------------------
class cvrnBundle_c;
class cvrnActiveKeyEntry_c;
class cvrnChannel_c;
class cvrnMcastChannel_c;
class cvrnMcastBundle_c;
class keyToolKey_c;
class keyToolManager_c;

/**
   Channel Buffer Class.
   Channel buffers are objects used to buffer incoming data destined for
   keys from various channels.
  */
class cvrnChannelBuffer_c {
public:


	globus_thread_t dispatchThread;

	/** STANDARD means TCP/UDP, MCAST is mcast.
	    This is to tell which form of connection this channel buffer belongs.
	    Depending on this type the parentChannel and parentMcastChannel
	    are used to point to the channels that are used to deliver data
	    for this buffer.
	    A better implementation is to have an mcast channel buffer and
	    a standard channel buffer subclass off a base channel buffer class.
	    */
	enum channelType_t {STANDARD, MCAST};

	/// Determine if the initial synch of a link has been done.
	int initialSynchDone;

	/** This is set to 1 if a remoteRequest call is done with CAVERN_link_c.
	It is set to 0 when the request is answered either by incoming data
	or a status message.
	*/
	int remoteRequestPending;

	/** This is the time out time set for remoteRequest calls.
	    When this timeout occurs, a callback is fired to the user.
	*/
	int remoteRequestTimeoutTime;

	// Flag used to block a remoteRequest call.
	int remoteRequestBlocking;
	CAVERN_irbLink_c::status_t remoteRequestStatus;

	globus_mutex_t remoteRequestPendingMutex;
	void lockRemoteRequestPending();
	void unLockRemoteRequestPending();

	globus_mutex_t channelBufferMutex;
	void lockChannelBuffer();
	void unLockChannelBuffer();

	/** ID of this channel buffer in the bundle's channel buffer list.
	    It is also the ID used to refer to channel buffers remotely.
	    */
	md5Key_c channelBufferId;

	/** ID of the remote channel buffer that this channel buffer is tied.
	  */
	md5Key_c remoteChannelBufferId;

	/** ID of this channel buffer in the active key database's channel
	    buffer list. This integer is the same as the bundleId
	    */
	int channelBufferIdInActiveKey;

	/// Type of parent channel- either standard (tcp / udp) or mcast.
	channelType_t parentChannelType;

	/// Channel that this buffer belongs.
	cvrnChannel_c *parentChannel;
	cvrnMcastChannel_c *parentMcastChannel;

	/// Active key entry.
	cvrnActiveKeyEntry_c *activeKeyEntry;

	/// The actual buffer.
	keyToolKey_c *bufferKey;

	/// Status of a channel buffer.
	enum status_t {
		CHANNEL_BUFFER_NOT_READY,
		CHANNEL_BUFFER_READY
	};
	status_t channelBufferReady;

	
	/// Update type of the link tied to this buffer (active or passive).
	CAVERN_linkAttrib_c :: CAVERN_update_t updateType;

	/// Initial and subsequent data synch types for the link tied to this buffer.
	CAVERN_linkAttrib_c :: CAVERN_synch_t initialSynchType, subsequentSynchType;

	/// Base constructor.
	cvrnChannelBuffer_c();

	~cvrnChannelBuffer_c();
	/// Initialize channel buffer.
	void  init(md5Key_c key,
		   cvrnActiveKeyEntry_c *theActiveKey, cvrnBundle_c *theBundle,
		   cvrnChannel_c *theChannel, cvrnMcastChannel_c *theMcastChannel,
		   CAVERN_linkAttrib_c *linkAttr, channelType_t chanType);

	/// Initialize mcast.
	void  initMcast(md5Key_c key, 
			cvrnActiveKeyEntry_c *theActiveKey, 
			cvrnMcastChannel_c *theMcastChannel);


};

/**
   This class holds the Nexus startpoint and endpoint pairs to allow 2 way communication
between a local and a remote channel. Channels may communicate over UDP or TCP.
*/
class cvrnChannel_c {

public:

	/// Ptr to the bundle that holds this channel.
	cvrnBundle_c *parentBundle;

	/// This local channel's Id
	int channelId;

	/// Remote channel's Id.
	int remoteChannelId;

	/// Startpoint and endpoint pairs for 2 way channel.
	globus_nexus_endpoint_t endpoint;
	globus_nexus_endpointattr_t endpointAttr;
	globus_nexus_startpoint_t startpoint;

	/// Not used.
	//	globus_nexus_mutexattr_t *mutexAttr;
	//	globus_nexus_mutex_t *mutex;

	/// Status values of channel.
	enum status_t {
		CHANNEL_NOT_READY, CHANNEL_READY, CHANNEL_CONNECT_FAILED,
		CHANNEL_CONNECT_TIMEOUT, CHANNEL_DEFUNCT, CHANNEL_FUNCTIONING
	};

	/// Type of channel, TCP or UDP.
	const int CVRN_CHANNEL_TCP = 1;
	const int CVRN_CHANNEL_UDP = 2;

	/// Channel type: either TCP or  UDP.
	int channelType;

	/// Ready for communication.
	status_t channelReady;

	/// Set to CHANNEL_FUNCTIONING by default. Set to CHANNEL_DEFUNCT
	/// if a Send RSR fails.
	status_t channelHealth;

	/// Constructor.
	cvrnChannel_c ();

	~cvrnChannel_c();

	/** This is a back pointer to the CAVERN_irbChannel_c that created this
	channel. If this channel was created dynamically by a remote request then
	this field is NULL. This ptr is primarily used to trace back to the
	CAVERN_irbChannel_c object that created this channel so that tigger callbacks
	can be done for the user.
	*/
	CAVERN_irbChannel_c *irbChannelCreator;

};

/** Class for a bundle. A bundle is set up for every independent remote client
we communicate with. So one bundle is dedicated to each client. And each bundle
can hold multiple channels with differing communication qualities.
The channel buffer list is a hash table of pointers to channel buffer objects.
In the bundle, the channel buffers are used to cache incoming data from multiple
remote sources.  There is also a channel buffer list in the active key manager. This
channel buffer list is bound to each key so that when a key is updated we know
which channel's to send data back to.
*/

class cvrnBundle_c {
public:

	/// Status' of a bundle.
	enum status_t {
		OK, FAILED,
		BUNDLE_READY, BUNDLE_NOT_READY, BUNDLE_FUNCTIONING,
		BUNDLE_DEFUNCT
	};

	/* Hash table of incoming data buffers assigned to the channels.
	   Hence there can only be a unique buffer per key per bundle.
	   This is so that you cannot link a key twice over 2 different
	   channels on the same bundle.
	   */
	_support_hashDict<cvrnChannelBuffer_c *, md5Key_c> *channelBufferList;

	/// Insert a channel buffer into the channel buffer list.
	void insertChannelBuffer(md5Key_c key, cvrnChannelBuffer_c *chanBuf, int cvrnLock = CVRN_NOLOCK);

	/** Remove a channel buffer from the channel buffer list. This returns a ptr to the hashed
	channel buffer but does not delete it. You have to delete it manually.
	*/
	void removeChannelBuffer(md5Key_c key, cvrnChannelBuffer_c *&chanBuf, int cvrnLock = CVRN_NOLOCK);

	/// Look for a channel buffer in the channel buffer list.
	void findChannelBuffer(md5Key_c key, cvrnChannelBuffer_c *& chanbuf, int cvrnLock = CVRN_NOLOCK);

	/// This bundle's id.
	int bundleId;

	/** Bundle ID of the remote client bundle that this bundle is
	connected to.
	*/
	int remoteBundleId;

	/// Determines if a bundle is ready for use.
	status_t bundleReady;
	status_t bundleHealth;

	/* NOT NEEDED ANYMORE
	/// remote IRB attach-URL if bundle was locally initiated.
	char remoteIRBAttachURLIfBundleLocallyInitiated[cvrn_URL_SIZE];
	*/

	/** URL of remote IRB that this bundle is attached. This is
	    sent by a remote IRB to the local IRB. Next time an attempt
	    is made to connect to the same IRB, the bundle is reused
	    rather than duplicated unnecessarily.
	    */
	char remoteIRBURL[cvrn_URL_SIZE];

	/** This determines if the current bundle was initiated locally
	or created as result of a remote link.
	*/
	CAVERN_bool_t bundleLocallyInitiated;

	// This is set to 1 if initial outgoingHailingSP was created ok in
	// attachToHailingServer.
	int bundleInitialCreationStatus;

	/// Startpoint that points to remote EP for the hailing channel.
	globus_nexus_startpoint_t outgoingHailingSP;

	/// Channel database. This is a hash table of channels keyed on channel Id.
	_support_hashDict<cvrnChannel_c *, int> *channelDB;

	/// Find a channel in the channel database.
	void findChannel(int id, cvrnChannel_c *&chan, int cvrnLock = CVRN_NOLOCK);
	void findReadyFunctioningChannel(int id, cvrnChannel_c *&chan, int cvrnLock = CVRN_NOLOCK);
	void findFunctioningChannel(int id, cvrnChannel_c *&chan, int cvrnLock = CVRN_NOLOCK);

	void removeChannel(int id, cvrnChannel_c *&chan, int cvrnLock = CVRN_NOLOCK);

	/// Create a channel in the current bundle.
	cvrnChannel_c * makeChannelInBundle(int channelType);

	/// private support function.
	cvrnChannel_c *createChannel(int *id);

	/// Used to lock this bundle.
	globus_mutex_t bundleLockMutex;

	/// Lock this bundle.
	void lock();
	/// Unlock this bundle.
	void unlock();

	cvrnBundle_c();
	~cvrnBundle_c();


	/* IMPORTANT TODO FOR CRASH RECOVERY
	void DeleteAllChannels();
	*/

};

/** Each CAVERN IRB has 1 communication manager.
 */
class cvrnCommunicationManager_c {

public:
	enum status_t {OK, FAILED};

	/** A communication manager has 1 hailing channel EP. Each client
	  can have only 1 communication manager. Hence there is only 1 hailing channel
	  EP per client.
	  */
	globus_nexus_endpointattr_t incomingHailingEPAttr;
	globus_nexus_endpoint_t incomingHailingEP;

	/// Database of bundles.
	_support_hashDict<cvrnBundle_c *, int> *bundleDB;

	/// There is only 1 mcast bundle per IRB.
	cvrnMcastBundle_c *theOnlyMcastBundle;

	/// Initialize comm manager (especially bundledb).
	status_t init();

	// This is the first thing you do to establish future communications
	// with remote clients. Create a new local bundle and set up a hailing channel
	// between you and the remote client.
	cvrnBundle_c *makeBundleAndHailingChannel(char *attach_url, status_t *status);

	/// Create a new bundle in the bundle database.
	cvrnBundle_c *createBundle(int *id);


	// This performs a globus_nexus_attach to the remote URL. This is called by
	// makeBundleAndHailingChannel.
	status_t attachToHailingServer(char *attach_url, cvrnBundle_c *theBundle);

	/// Look for a locally  initiated bundle given the URL.
	void findLocallyInitiatedBundle(char *attach_url, cvrnBundle_c *&, int cvrnLock = CVRN_NOLOCK);

	/// Look for a bundle given the URL.
	void findBundle(char *attach_url, cvrnBundle_c *&, int cvrnLock = CVRN_NOLOCK);
	void findReadyFunctioningBundle(char *attach_url, cvrnBundle_c *&, int cvrnLock = CVRN_NOLOCK);
	void findFunctioningBundle(char *attach_url, cvrnBundle_c *&, int cvrnLock = CVRN_NOLOCK);

	/// Look for a bundle given the ID.
	void findBundle(int id, cvrnBundle_c *&, int cvrnLock = CVRN_NOLOCK);

	void findReadyFunctioningBundle(int id, cvrnBundle_c *&entry, int cvrnLock = CVRN_NOLOCK);

	void findFunctioningBundle(int id, cvrnBundle_c *&entry, int cvrnLock = CVRN_NOLOCK);

	/// Remove a bundle given the ID. Returns the bundle object for manual deletion.
	void removeBundle(int id, cvrnBundle_c *&, int cvrnLock = CVRN_NOLOCK);

};


//----------------------------------------------------------------------

// Hailing channel handlers 

#define REMOTE_BUNDLE_CREATE_HANDLER_ID        0
#define REMOTE_BUNDLE_CREATE_REPLY_HANDLER_ID       1
#define HAIL_RECEIVE_DATA_HANDLER_ID 2
#define CREATE_NEW_CHANNEL_HANDLER_ID 3
#define CREATE_REPLY_NEW_CHANNEL_HANDLER_ID 4
#define MAKE_LINK_HANDLER_ID 5
#define REPLY_MAKE_LINK_HANDLER_ID 6
#define REPLY_MAKE_LINK_FAILURE_HANDLER_ID 7
#define REMOTE_REQUEST_HANDLER_ID 8
#define REMOTE_REQUEST_REPLY_HANDLER_ID 9
#define REMOTE_COMMIT_RPC_HANDLER_ID 10
#define REMOTE_UNLINK_RPC_HANDLER_ID 11
#define DELETE_REMOTE_STANDARD_CHANNEL_RPC_HANDLER_ID 12
#define DELETE_REMOTE_STANDARD_BUNDLE_RPC_HANDLER_ID 13
#define HEALTH_PULSE_HANDLER_ID 14
#define ACK_STANDARD_DATA_RECEIPT_HANDLER_ID 15

// Status messages from various handler calls.

#define WAITING_FOR_MAKE_LINK 0
#define MAKE_LINK_STATUS_OK 1
#define MAKE_LINK_STATUS_FAIL 2
//#define MAKE_LINK_STATUS_KEY_NOT_EXIST 3
#define MAKE_LINK_STATUS_MAKE_KEY_FAILED 3
#define MAKE_LINK_STATUS_COMM_FAIL 4
#define MAKE_LINK_STATUS_ALREADY_LINKED 5

#define CVRN_REMOTE_REQUEST_STATUS_FAIL 1
#define CVRN_REMOTE_REQUEST_STATUS_NO_DOWNLOAD_NEEDED 2
#define CVRN_REMOTE_REQUEST_STATUS_OK 3


/* A response object is like a "token" that is
passed from a client to a server and then returned back to a
client. This token is used to stuff an address of some storage
location (possibly an object) that needs to be accessed
when the server responds back to the client. In some sense it
serves as a "naming" scheme like in named threads in pthreads.
A response object is usually only 4 bytes in size (sizeof(int*))
but to be safe and to save me from having to dynamically allocate and
delete it, we make it 16 bytes in size.
*/
#define RESPONSE_OBJECT_TOO_BIG_SIZE 16

// Nexus handlers for each of the different hailing channel message types.

void cvrn_NX_bundleCreationHandler(globus_nexus_endpoint_t *endpoint, 
				  globus_nexus_buffer_t *buffer, 
				  globus_bool_t is_non_threaded_handler);
void cvrn_NX_bundleCreationReplyHandler(globus_nexus_endpoint_t *endpoint, 
				       globus_nexus_buffer_t *buffer,
				       globus_bool_t is_non_threaded_handler);
void cvrn_NX_hailReceiveDataHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);
void cvrn_NX_createNewChannelHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);
void cvrn_NX_createReplyNewChannelHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);
void cvrn_NX_channelIncomingDataHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

void cvrn_NX_makeLinkHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

void cvrn_NX_replyMakeLinkHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

void cvrn_NX_replyMakeLinkFailureHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

void cvrn_NX_remoteRequestHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

void cvrn_NX_remoteRequestReplyHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

void cvrn_NX_remoteUnlinkHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

void cvrn_NX_remoteCommitRPCHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

void cvrn_NX_remoteUnlinkRPCHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

void cvrn_NX_deleteRemoteStandardChannelRPCHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

void cvrn_NX_deleteRemoteStandardBundleRPCHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

void cvrn_NX_acknowledgeStandardDataReceiptHandler(globus_nexus_endpoint_t *endpoint,  //@DEF@
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

extern globus_nexus_handler_t cvrn_NX_hailingHandlers[];
int cvrnRemoteCommitRPC(cvrnBundle_c *theBundle, cvrnChannelBuffer_c *chanBuffer);
int cvrnRemoteUnLinkRPC(cvrnBundle_c *theBundle, cvrnChannelBuffer_c *chanBuffer);
int cvrnDeleteRemoteStandardChannelRPC(cvrnBundle_c *theBundle, int channelId);
int cvrnDeleteRemoteStandardBundleRPC(cvrnBundle_c *theBundle);

// Channel handlers-----------------------------------------------------------

#define INCOMING_CHANNEL_DATA_HANDLER_ID 0
#define CHANNEL_STANDARD_DATA_HANDLER_ID 1
#define MCAST_CHANNEL_DATA_HANDLER_ID 2

// Junna, added non-threaded handler id
#define NON_THREADED_CHANNEL_STANDARD_DATA_HANDLER_ID 3
#define NON_THREADED_MCAST_CHANNEL_DATA_HANDLER_ID 4

// Handler for incoming data on a channel. This is present only for initial
// debugging purposes.
void cvrn_NX_channelIncomingDataHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

// Handler for STANDARD_MESSAGE on a channel.
void cvrn_NX_channelStandardDataHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

// Junna, added non-threaded handler
void cvrn_NX_non_threaded_channelStandardDataHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

void cvrn_NX_healthPulseHandler(globus_nexus_endpoint_t *endpoint, 
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

void cvrn_NX_channelMcastDataHandler(globus_nexus_endpoint_t *endpoint,
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

// Junna, added non-threaded handler
void cvrn_NX_non_threaded_channelMcastDataHandler(globus_nexus_endpoint_t *endpoint,
				   globus_nexus_buffer_t *buffer,
				   globus_bool_t is_non_threaded_handler);

// Handlers for channel events. Still need more events.
extern globus_nexus_handler_t cvrn_NX_channelHandlers[];

//---------------------------------------------------------------------------

// Initialize the communication manager.
cvrnCommunicationManager_c * cvrnCommunicationManagerInit();

// Init Nexus.
void cvrnInitNexus(int *argc, char ***argv);

// Second half of communication manager init that is done after Nexus inits.
int cvrnCommunicationManagerPostInit();

// Allow this client to accept remote attachments.
int cvrnAllowHailingAttachments(unsigned short *attach_port, char *incomingURL);

// This is a callback function that is called when an incoming attachment occurs.
int cvrnHailingAttachRequested(void *arg, char* url, globus_nexus_startpoint_t *sp);

// Initialize the ID generator.
void cvrnInitNewIDGenerator();

// Generate a new ID for either a bundle id or a channel id.
int cvrnGenerateNewID();

/** This is called by CAVERN_irb_c::define to define a new remote key. */
class CAVERN_irbKeyId_c;

//----------------------------------------------------------------------

/* Multicast channel class. */
class cvrnMcastChannel_c {
public:

	cvrnMcastChannel_c();

	enum status_t  {
		CHANNEL_NOT_READY, CHANNEL_READY, CHANNEL_DEFUNCT, CHANNEL_FUNCTIONING
	};

	unsigned long mcastGroup;
	unsigned short mcastPort;

	unsigned long mcastChannelId; // Packs both mcastGroup and mcastPort

	_support_hashDict<cvrnChannelBuffer_c *, md5Key_c> * channelBufferList;
	status_t channelReady;
	status_t channelHealth;

	/** This is a back pointer to the CAVERN_irbChannel_c that created this
	channel. If this channel was created dynamically by a remote request then
	this field is NULL. This ptr is primarily used to trace back to the
	CAVERN_irbChannel_c object that created this channel so that tigger callbacks
	can be done for the user.
	*/
	CAVERN_irbMcastChannel_c *irbChannelCreator;

	globus_nexus_startpoint_t mcastSP;
	globus_nexus_endpoint_t mcastEP;

	void findChannelBuffer(md5Key_c key, cvrnChannelBuffer_c *&chanbuf,
					     int cvrnLock);

	void insertMcastChannelBuffer(md5Key_c key,
				      cvrnChannelBuffer_c * chanbuf, int cvrnLock);

};

class cvrnMcastBundle_c {
public:
	// Mcast Channel database. - may want to key on mcastgroup.
	_support_hashDict<cvrnMcastChannel_c *, unsigned long> *mcastChannelDB;
	cvrnMcastBundle_c();

	cvrnMcastChannel_c *makeMcastChannelInBundle(char* mcastAddr, unsigned short mcastPort, int ttl=128);

	void removeChannelBuffer(unsigned long channelId, md5Key_c chanBufId,
				 cvrnChannelBuffer_c *&chanBuff,
				 int cvrnLock = CVRN_NOLOCK);
	void findMcastChannel(unsigned long chanId, cvrnMcastChannel_c *&theChan,
			      int cvrnLock = CVRN_NOLOCK);


};

// Nexus put for md5 key
void cvrn_globus_nexus_put_md5(globus_nexus_buffer_t *buffer, md5Key_c key);

// Grab the time in seconds.
double cvrnGetTimeInSecs();

int removeChannelBufferFromMcastChannelAndActiveKeyManager(unsigned long mcastChannelId,
							   md5Key_c channelBufferId);

int removeChannelBufferFromBundleAndActiveKeyManager(int  bundleId, md5Key_c channelBufferId);
int removeChannelBufferFromBundleAndActiveKeyManager(cvrnBundle_c *theBundle, md5Key_c channelBufferId);


int purgeStandardChannelAndItsChannelBuffers(int bundleId, int channelId, int doRemoteToo);
int purgeStandardChannelAndItsChannelBuffers(cvrnBundle_c *theBundle, int channelId, int doRemoteToo);
int purgeStandardBundle(int bundleId, int doRemoteToo);

void cvrnSendBundleHealthPulse(cvrnBundle_c *theBundle);

#endif


