/* Author: Jason Leigh
 * (C) 1997 Electronic Visualization Laboratory - Univ Illinois at Chicago
 */
#ifndef _CVRN_DB
#define _CVRN_DB

#include "cvrn_nexus.h"
#include "cvrn_base.h++"
#include "_support_hashDict.h++"
#include "cvrn_comm.h++"

#include "CAVERN_includes.h++"
#include "CAVERN_irbKey_c.h++"

// This tells dispatch to check and only dispatch data to active keys.
#define _CVRN_DB_DISPATCH_CHECK_ACTIVE 1

// This tells dispatch to dispatch data to the key regardless of active or passive.
#define _CVRN_DB_DISPATCH_NO_CHECK_ACTIVE 0

class keyToolKey_c;
class keyToolManager_c;

const int _cvrnMaxActiveKeys = 50;
class cvrnActiveKeyEntry_c;

class cvrnChannelBuffer_c;

/** Manager for active keys */
class cvrnActiveKeyManager_c {
public:
	enum status_t {
		FAILED, OK
	};

	enum loadMode_t {
		CREATE_IF_NOT_EXIST,
		FAIL_IF_NOT_EXIST
	};

	keyToolManager_c *keyToolManager;
	_support_hashDict<cvrnActiveKeyEntry_c *, md5Key_c> *activeKeyDB;
	status_t init();

	int existsActiveKey(md5Key_c akey, int cvrnLock = CVRN_NOLOCK);
	void findActiveKey(md5Key_c akey, cvrnActiveKeyEntry_c *&,  int cvrnLock = CVRN_NOLOCK);

	// Increase the reference count of the active key entry.
	void ref(md5Key_c akey, int cvrnLock = CVRN_NOLOCK);

	// Decrease the reference count of the active key entry. If entry's ref
	// count becomes zero then the active key entry is deleted. It is assumed that
	// by this time the channel buffer list in the key entry would have been
	// empty so that there would not be any channel buffers that need to be
	// deleted.
	void deref(md5Key_c akey, int cvrnLock = CVRN_NOLOCK);

	// Load a key from a keytool manager creating the proper active key entry etc..
	void loadKeyFromKeyTool(CAVERN_irbKeyId_c *keyInfo, cvrnActiveKeyEntry_c *&,
				loadMode_t loadMode = CREATE_IF_NOT_EXIST,
				int cvrnLock = CVRN_NOLOCK);

	// Dispatch data in active key trueKey to all channel buffers except the specified
	// by exceptChanBuf (usually the originator of the message, or NULL if the local
	// client did the dispatch. This will check to see what active/passive update is set to.
	// The checkActiveParam field gets either: 
	// _CVRN_DB_DISPATCH_CHECK_ACTIVE or _CVRN_DB_DISPATCH_NO_CHECK_ACTIVE
	// Junna, added parameter triggerType to dispatch()
	void dispatch(cvrnActiveKeyEntry_c *entry, cvrnChannelBuffer_c *exceptChanBuf, int checkActiveParam, int
waitForAcknowledge, CAVERN_irbKey_c::CAVERN_irbKeyTrigger_t triggerType);

	// Deliver the data in the active key pointed to by the channel buffer
	// over the comm channel pointed to by channel buffer.
	// Delivery occurs regardless of what active/passive update was set to.
	// Junna, added parameter triggerType to deliver()
	void deliver(cvrnChannelBuffer_c *chanBuf, int forcedDeliverRegardlessOfSynchType, int
waitForAcknowledge, CAVERN_irbKey_c::CAVERN_irbKeyTrigger_t triggerType);
};

/** An active key entry */
class cvrnActiveKeyEntry_c {
public:

	int referenceCount;

	void ref();
	void deref();

	////////////////////////////////////////////////////////
	// For handling user trigger callbacks.
	void (*userCallback)(CAVERN_irbKey_c::CAVERN_irbKeyEvent_t event,
			     CAVERN_irbKey_c *theIrbKey,
			     void* userData);
	void *userCallbackData;

	// This is a ptr to the original CAVERN_irbKey_c that
	// set the trigger.
	CAVERN_irbKey_c *userIRBKeyForCallback;
	////////////////////////////////////////////////////////


	nexus_mutex_t simultaneousSwapMutex;
	nexus_mutex_t trueKeyMutex;
	nexus_mutex_t referenceCountMutex;
	nexus_mutex_t shadowBufferMutex;

	// Apply this lock before using the shadow buffer below.
	// This is solely used in CAVERN_irbKey::put
	void lockShadowBuffer();

	// Apply this lock before attempting to do a CAVERN_irbKey::put
	// or any operation that will involve performing a keytool swap.
	// It also makes sure that user callbacks are done atomically.
	void lockSimultaneousSwap();

	// Lock this key if accessing the data element of truekey
	void lockTrueKey();

	void unLockShadowBuffer();
	void unLockSimultaneousSwap();
	void unLockTrueKey();

	/// Key holding what is considered the current true data of a CAVERN key.
	keyToolKey_c *trueKey;

	double trueKeyTimeStamp;

	/// Return the checksum of the data in the true key.
	md5Key_c getDataCheckSum(int cvrnLock = CVRN_NOLOCK);

	/// Buffer for local client to make updates to.
	keyToolKey_c *localShadowBufferKey;

	/// Array of data buffers representing external subscribers, assigned to this key.
	_support_hashDict<cvrnChannelBuffer_c *, int> *channelBufferList;
	void insertChannelBuffer(int key, cvrnChannelBuffer_c * chanbuf, int cvrnLock = CVRN_NOLOCK);

	/// Remove a channel buffer from the channel buffer list. Return a ptr to the
	/// channel buffer but do not delete it. You have to delete it manually.
	void removeChannelBuffer(int key, cvrnChannelBuffer_c *&chanbuf, int cvrnLock = CVRN_NOLOCK);
	
	/// Init active key.
	int init(cvrnActiveKeyManager_c *activeKeyManager, CAVERN_irbKeyId_c *newKeyInfo);

	cvrnActiveKeyEntry_c();

};

cvrnActiveKeyManager_c* cvrnActiveKeyManagerInit();
int cvrnActiveKeyManagerPostInit(CAVERN_initAttrib_c *initAttr);

//----------------------------------------------------------------------


#endif
