/* Author: Jason Leigh
 * (C) 1997 Electronic Visualization Laboratory - Univ Illinois at Chicago
 */
#include "_support_hashDict.h++"


template <class valueType, class keyType>
_support_hashDict<valueType,keyType>::_support_hashDict( int entries ) {

#ifdef NEXUS_THREAD_SAFE
		nexus_mutex_init(&mutex,NULL);
#endif
#ifdef PTHREAD_SAFE
		pthread_mutex_init(&mutex,NULL);
#endif

	tableSize=entries;
	buckets=new _support_hashDictEntry<valueType,keyType> *[tableSize];
	for (int i = 0; i < tableSize; i++) {
		buckets[i] = NULL;
	}
	numberOfItems = 0;
}

template <class valueType, class keyType>
_support_hashDict<valueType,keyType>::~_support_hashDict() {
	clear();
	delete buckets;

#ifdef NEXUS_THREAD_SAFE
		nexus_mutex_destroy(&mutex);
#endif
#ifdef PTHREAD_SAFE
		pthread_mutex_destroy(&mutex);
#endif

}

template <class valueType, class keyType>
void
_support_hashDict<valueType,keyType>::clear() {
	int i;
	_support_hashDictEntry<valueType,keyType> *entry, *nextEntry;
  
	for (i = 0; i < tableSize; i++) {
		for (entry = buckets[i]; entry != NULL; entry = nextEntry) {
			nextEntry = entry->next;
			delete entry;
		}
		buckets[i] = NULL;
	}
	numberOfItems = 0;
}




template <class valueType, class keyType>
int
_support_hashDict<valueType,keyType>::enter(keyType key, valueType value) {
_support_hashDictEntry<valueType,keyType>           *&entry = findEntry(key);
  
	if (entry == NULL) {
		entry = new _support_hashDictEntry<valueType,keyType>(key, value);
		entry->next = NULL;
		numberOfItems++;
		return 1;
	}
	else {
		entry->value = value;
		return 0;
	}
}

template <class valueType, class keyType>
int
_support_hashDict<valueType,keyType>::find(keyType key, valueType &value) const {

	_support_hashDictEntry<valueType,keyType>*&entry = findEntry(key);
  
	if (entry == NULL) {
		value = NULL;
		return 0;
	}
	else {
		value = entry->value;
		return 1;
	}
}

template <class valueType, class keyType>
_support_hashDictEntry<valueType,keyType> *&
_support_hashDict<valueType,keyType>::findEntry(keyType key) const {
	_support_hashDictEntry<valueType,keyType>           **entry;

	/*  
	int *keyFront = (int*) (key.key);

	int copyFront = *keyFront;

	// Make sure keyfront is positive.
	if (copyFront < 0) copyFront = -copyFront;
	*/

	int copyFront = key;

	entry = &buckets[copyFront % tableSize];

	while (*entry != NULL) {
		if ((*entry)->key == key) {
			break;
		}
		entry = &(*entry)->next;
	}
	return *entry;
}

template <class valueType, class keyType>
int
_support_hashDict<valueType,keyType>::remove(keyType key) {
	_support_hashDictEntry<valueType,keyType>           *&entry = findEntry(key);
	_support_hashDictEntry<valueType,keyType>           *tmp;
  
	if (entry == NULL) {
		return 0;
	}
	else {
		tmp = entry;
		entry = entry->next;
		delete tmp;
		numberOfItems--;
		return 1;
	}
}


template <class valueType, class keyType>
valueType*
_support_hashDict<valueType,keyType>::buildListOfEntries(int *number)
{
	int i;
	_support_hashDictEntry<valueType,keyType> *entry;

	valueType *list;

	int numItems = getNumItems();

	if (numItems == 0) {
		*number = 0;
		return NULL;
	}

	list = new valueType[numItems];
	if (list == NULL) {
		*number = 0;
		return NULL;
	}

	*number = numItems;

	int index = 0;

	// Call rtn for each entry in dict
	for (i = 0; i < tableSize; i++) {

		for (entry = buckets[i]; entry != NULL; entry = entry->next) {
			list[index] = entry->value;
			index++;
		}
	}
	return list;
}



// Calls given routine (passing value) for each entry in dictionary.
// The order of entries is not guaranteed to mean anything.
template <class valueType, class keyType>
void
_support_hashDict<valueType,keyType>::applyToAll(int (*rtn)(keyType key, valueType value) ) {
	int i;
	_support_hashDictEntry<valueType,keyType> *entry;

	// Call rtn for each entry in dict
	for (i = 0; i < tableSize; i++) {

		for (entry = buckets[i]; entry != NULL; entry = entry->next) {
			if (((*rtn)(entry->key, entry->value)) == 0) return;
		}
	}
}

template <class valueType, class keyType>
void _support_hashDict<valueType,keyType>::showHashProfile()
{
	int i, count;
	_support_hashDictEntry<valueType,keyType> *entry;

	// Call rtn for each entry in dict
	for (i = 0; i < tableSize; i++) {
		count = 0;
		for (entry = buckets[i]; entry != NULL; entry = entry->next) {
			count++;
		}
		if (count)
		printf("%04d: %d\n",i,count);
	}

}


// Calls given routine (passing value,data) for each entry in dictionary.
// The order of entries is not guaranteed to mean anything.
template <class valueType, class keyType>
void
_support_hashDict<valueType,keyType>::applyToAll(int (*rtn)(keyType key, valueType value,
							    void *data1, void *data2,
							    void *data3, void *&retdata),
						 void *data1, void *data2,
						 void *data3, void *&retdata) {
	int i;
	_support_hashDictEntry<valueType,keyType> *entry;

	// Call rtn for each entry in dict
	for (i = 0; i < tableSize; i++) {

		for (entry = buckets[i]; entry != NULL; entry = entry->next)
			if(((*rtn)(entry->key, entry->value, data1,data2,data3,retdata)) == 0) return;
	}
}


