/* Author: Jason Leigh
 * (C) 1997 Electronic Visualization Laboratory - Univ Illinois at Chicago
 */
#ifndef _SUPPORT_HASHDICT_
#define _SUPPORT_HASHDICT_

#include <stdlib.h>
//#include "_support_hash.h++"
//#include "md5Key_c.h++"

//define NEXUS_THREAD_SAFE
//#define PTHREAD_SAFE
//#define NO_THREAD

#ifdef NEXUS_THREAD_SAFE
#include "globus_nexus.h"
#endif

#ifdef PTHREAD_SAFE
#include <pthread.h>
#endif

#ifdef NEXUS_THREAD_SAFE
#define _SUPPORT_HASHDICT_LOCK	globus_mutex_lock(&mutex);
#define _SUPPORT_HASHDICT_UNLOCK globus_mutex_unlock(&mutex);
#endif
#ifdef PTHREAD_SAFE
#define _SUPPORT_HASHDICT_LOCK	pthread_mutex_lock(&mutex);
#define _SUPPORT_HASHDICT_UNLOCK pthread_mutex_unlock(&mutex);
#endif
#ifdef NO_THREAD
#define _SUPPORT_HASHDICT_LOCK
#define _SUPPORT_HASHDICT_UNLOCK
#endif


template <class valueType, class keyType>
class _support_hashDict;

/*  A dictionary is stored as a collection of entries, each of which
 *  is an _support_hashDictEntry. It contains the key and value for the entry and
 *  a link to create lists of entries. 
 */
template <class valueType, class keyType>
class _support_hashDictEntry {

private:
	keyType		key;
	valueType		value;
	_support_hashDictEntry<valueType,keyType> *	next;

	_support_hashDictEntry(keyType k, valueType v)	{ key = k; value = v; };

	friend class _support_hashDict<valueType,keyType>;
};



/* _support_hashDict: a dictionary mapping (keyType) keys to 
 * (valueType) data pointers.
 * Whatever keyType is, it needs to have a type conversion mechanism to integer.
 * So if you are planning keyType to be some kind of object then that object's class
 * needs an operator int() converter in its class.
 */
template <class valueType, class keyType>
class _support_hashDict {

public:
	_support_hashDict(int entries = 1021);
	~_support_hashDict();


	/// Manually lock this array object. (not recommended)
	void lock() {
		_SUPPORT_HASHDICT_LOCK
	}

	/// Manually unlock this array object. (not recommended)
	void unlock() {
		_SUPPORT_HASHDICT_UNLOCK
	}

	// Calls given routine (passing value) for each entry in dictionary.
	//  The order of entries is not guaranteed to mean anything.
	// If routine returns 0 during an iteration the apply exits.
	// This can be used as an escape hatch.
	void applyToAll(int (*rtn)(keyType key, valueType value) );

	void showHashProfile();

	// This builds a linear array of entries out of the hash dictionary.
	// You must delete the list when done.
	valueType * buildListOfEntries(int *number);

	// Calls the given routine (passing value,data) for each entry in dictionary.
	//  The order of entries is not guaranteed to mean anything.
	// If routine returns 0 during an iteration the applyToAll exits. This can
	// be used as an escape hatch.
	// void *&retdata can be used to pass data into as well as out from the iteration.
	// Note that data passed in must first be type casted to void* to a separate
	// temporary ptr before the ptr is passed to the member function.
	// e.g. void *f;
	// f = (void*) data;
	// applyToAll(xxx,xxx,xxx,xxx, f);
	void applyToAll(int (*rtn)(keyType key, valueType value,
				   void *data1, void *data2, void *data3, void *&retdata),
			void *data1, void *data2, void *data3, void *&retdata );
	// Removes all entries from dictionary.
	void clear();

	/* Enters a key,value pair into the dictionary. Overwrites entry and
	 * returns FALSE if key already exists.
	 */
	int enter(keyType key, valueType value);

	/* Finds entry with given key, setting value to point to value.
	 * Returns FALSE if no such entry.
	 */
	int find(keyType key, valueType &value) const;

	// Removes the entry with the given key. Returns FALSE if no such entry
	int remove(keyType key);

	_support_hashDictEntry<valueType,keyType> *&findEntry(keyType key) const;

	int getNumItems(){return numberOfItems;}
private:

	int numberOfItems;

	// Entries in table
	int tableSize;

	// Entries are stored as an external hash table of _support_hashDictEntry instances.
	_support_hashDictEntry<valueType,keyType> **buckets;

#ifdef NEXUS_THREAD_SAFE
	globus_mutex_t mutex;
#endif
#ifdef PTHREAD_SAFE
	pthread_mutex_t mutex;
#endif


};



#endif // _SUPPORT_HASHDICT_ 
