#include "CAVERN_includes.h++"

/** CAVERN direct support for TCP/IP. It is not recommended that you use this
except to connect to clients and servers that do not use IRBs (legacy systems).
TCP is a reliable streaming protocol.

This object is instantiated by calling: CAVERN_irb_c::makeTCPKey().
Once you have this key you can choose to make it a client connection or a server.

To make it a client, call the connect()  member function. Then you may call
put or get as needed.

To make it a server, call the accept() member function. Each time a remote
client connects to you a user-defined callback will be fired and a new TCP
key is allocated. You can then use the new key to communicate with the connection.
You should not do calls to get and put with the server key. You should only
do it with the keys generated from the accept() callback.

Finally by using a trigger you can be asynchronously notified of when a key
has new incoming data.


*/
class CAVERN_TCPKey_c : public CAVERN_key_c {

public:

	/// Events that might occur in TCP connections.
	enum CAVERN_TCP_triggerEvent_t  {

		/// New data has arrived.
		NEW_DATA_ARRIVED,

		/// A connection has closed.
		CONNECTION_CLOSED_EVENT
	};

	/// Status codes.
	enum statusCodes_t  {

		/// Operation successful.
		OK = 0,
		/// Operation failed.
		FAILED = -1,
		/// QoS negotiation made.
		NEGOTIATED = -2,

		/** For nonblocking socket calls this is the status if
		a read is attempted on a socket with no data available to be read.
		*/
		NON_BLOCKING_NO_DATA_TO_READ = -3,

		/// Connection closed during get/put call.
		CONNECTION_CLOSED = -4,

		/// Reading or writing data timedout.
		TIME_OUT = -5
	};

	/// Used in setTimeOut() to turn off timeouts during WEB download.
	const int NEVER_TIME_OUT = -1;

	/// Blocking/Non-blocking attribute type.
	enum CAVERN_blocking_t  {
		/// Make an operation blocking.
		BLOCKING,
		/// Make an operation non-blocking (ie return immediately)
		NON_BLOCKING
	};

	/** If connect is called then assuming role of a client connection.
	  and hence connect to a remote server. An accept call assumes you are
	  taking the role of a server. You can not perform a connect() and
	  an accept() together on a single key.
	  */
	statusCodes_t connect(char *ip, unsigned short port);

	/** Researve Quality of Service
	@param Desired qos attributes. Also returns with contract.
	@return OK if got contract ; NEGOTIATED if had to negotiate for lower ;
	FAILED if completely failed.
	*/
	statusCodes_t reserveQoS(CAVERN_qosAttrib_c *qos);

	/** Accept incoming TCP connections.
	  If accept is called then assuming role of a server.

	  @param callback gets called if a new incoming connection is established.

	  @param newConnectionKey is the key that points to the newly
	  established connection.  newConnectionKey is the dynamically
	  generated key for the new connection.

	  Callback's thisKey parameter contains a pointer to the key that
	  accepted the incoming connection- namely this key.

	  @param userData is data that the user may pass when the callback is called.  */
	void accept(unsigned short port,
		   void (*callback)(CAVERN_TCPKey_c *thisKey,
				    CAVERN_TCPKey_c *newConnectionKey, void* userData),
		   void *userData);

	/** Trigger on TCP events.
	 */
	void trigger(void (*callback)(CAVERN_TCP_triggerEvent_t event,
				      CAVERN_TCPKey_c *thisKey,
				      void* userData),
			void* userData);

	/** Get data from socket.
	@return status:
	Returns a status code or else status = UNIX errno.
	@see statusCodes_t

	@param buffer to fill with data.
	@param sizeOfBuffer specifies the number of bytes to put into buffer.
	It also returns the number of bytes successfully extracted.
	@param howto determines if the call blocks until a timeout period or is
	non blocking.
	@param timeOut is timeout period that get blocks waiting for data. If connection
	is non blocking then timeOut is the period it waits once it has started getting
	data. i.e. if a non blocking get is attempted when no data is available, it returns
	immediately and returns a status of -1.
	*/
	statusCodes_t get(char * buffer, int *sizeOfBuffer, CAVERN_blocking_t howto,
		int timeOut = NEVER_TIME_OUT);

	/** Send data over TCP.
	  @param buffer contains data to send.
	  @param sizeOfBuffer contains amount of data in buffer to send. Returns
	  the amount of data successfully sent.
	  @return status:
	  Returns a status code or else status = UNIX errno.
	  @see statusCodes_t
	  */
	statusCodes_t put(char *buffer, int *sizeOfBuffer);

	/// Return IP address of self.
	unsigned long getSelfIP();

	/// Return IP address of self as a string xxx.xxx.xxx.xxx.
	void getSelfIP(char *ip);

	/// Return port of self.
	unsigned short getSelfPort();

	/** Return IP address of remote connection. If this is a server object
	then this returns 0.
	*/
	unsigned long getRemoteIP();

	/// Return IP address of remote connection as a string xxx.xxx.xxx.xxx.
	void getRemoteIP(char *ip);

	/// Return port of remote connection.
	unsigned short getRemotePort();

	/// Disconnect socket connection and tell IRB to unmanage it.
	void disconnect();
};

