#include "CAVERN_includes.h++"

/** CAVERN direct support for UDP. It is not recommended that you use this
except to connect to clients and servers that do not use IRBs (legacy systems).
UDP is an unreliable packet protocol. Keep the packets small.

You can use this either as a client or a server like in TCP but remember UDP
is connectionless.

To open the connection as a client so that you can send data out to remote sites,
call open(). Then call setSendAddress(). Once this is done you can send a packet to
that address. You can at anytime change the send address by calling setSendAddress().

To open as a server call open(). Usually you want to call open with a specific address
so that the remote clients have an exact port to connect to. Once opened, simply
wait for incoming data. You can use the trigger to notify you when that happens.

For each piece of incoming data you can find out where they came
from. By calling copyReceiveAddressToSendAddress()
you can then send a message back to the original sender.

*/
class CAVERN_UDPKey_c : public CAVERN_key_c {
public:

	/// Status codes for UDP operations.
	enum statusCodes_t  {

		/// Operation succeeded.
		OK = 0,
		/// Operation failed.
		FAILED = -1,
		/// A lower QoS service  was negotiated.
		NEGOTIATED = -2,

		/** For nonblocking socket calls this is the status if
		a read is attempted on a socket with no data available to be read.
		*/
		NON_BLOCKING_NO_DATA_TO_READ = -3
	};

	/// Trigger Events for UDP.
	enum CAVERN_UDP_triggerEvent_t  {

		/// New data has arrived.
		NEW_DATA_ARRIVED
	};

	/** Open UDP socket */
	statusCodes_t open(int port=0); // port = 0 for dynamic port assignment

	/** Choose a destination to send data */
	statusCodes_t setSendAddress(char *ip, int port);

	/** Researve Quality of Service
	@param Desired qos attributes. Also returns with contract.
	@return OK if got contract ; NEGOTIATED if had to negotiate for lower ;
	FAILED if completely failed.
	*/
	statusCodes_t reserveQoS(CAVERN_qosAttrib_c *qos);

	/** Trigger on UDP events. There is currently only 1 event NEW_DATA_ARRIVED.
	  There is no CONNECTION_CLOSED since UDP is connectionless.
	 */
	void trigger(void (*callback)(CAVERN_UDP_triggerEvent_t event,
				      CAVERN_UDPKey_c *thisKey,
				      void* userData),
			void* userData);

	/** Get data.
	  @return status
	  Returns a status code or else status = UNIX errno.
	  @see statusCodes_t

	  @param buffer to fill with data.
	  @param sizeOfBuffer specifies the number of bytes to put into buffer.
	  It also returns the number of bytes successfully extracted.
	  */
	statusCodes_t get(char * buffer, int *sizeOfBuffer);

	/** Send data.
	  @param buffer contains data to send.
	  @param sizeOfBuffer contains amount of data in buffer to send. Returns
	  the amount of data successfully sent.
	  @return status:
	  Returns a status code or else status = UNIX errno.
	  @see statusCodes_t
	  */
	statusCodes_t put(char *buffer, int *sizeOfBuffer);

	/** Take an incoming packet's address and make this the outgoing address
	  so that you can send data directly back to them.
	  An alternative method is to instantiate a new UDP object and set the
	  destination using the setSendAddress call.
	  */
	void copyReceiveAddressToSendAddress();

	/// Get an incoming message's ip and port.
	//@{
	/// Get incoming message's ip.
	unsigned long getReceiveIP();
	/// Get incoming message's ip as a string.
	void getReceiveIP(char *ip);
	/// Get incoming message's port.
	unsigned short getReceivePort();
	//@}


	/// Get destination's ip and port.
	//@{
	/// Get destination ip.
	unsigned long getSendIP();
	/// Get destination ip as string.
	void getSendIP(char *ip);
	/// Get destination port.
	unsigned short getSendPort();
	//@}

	/// Get this key's ip and port.a
	//@{
	/// Get self ip.
	unsigned long getSelfIP();
	/// Get self ip  as string.
	void getSelfIP(char *ip);
	/// Get self port.
	unsigned short getSelfPort();
	//@}

	/// Close udp key and tell IRB to unmanage it.
	void close();
};
