#ifndef _MANAGED_ARRAY__
#define _MANAGED_ARRAY__

#include <stdlib.h>
#include <stdio.h>

//define NEXUS_THREAD_SAFE
//#define PTHREAD_SAFE
//#define NO_THREAD

#include "_support_hash.h++"

#ifdef NEXUS_THREAD_SAFE
#include "nexus.h"
#endif

#ifdef PTHREAD_SAFE
#include <pthread.h>
#endif

#ifdef NEXUS_THREAD_SAFE
#define _MANAGED_ARRAY_LOCK	nexus_mutex_lock(&mutex);
#define _MANAGED_ARRAY_UNLOCK nexus_mutex_unlock(&mutex);
#endif
#ifdef PTHREAD_SAFE
#define _MANAGED_ARRAY_LOCK	pthread_mutex_lock(&mutex);
#define _MANAGED_ARRAY_UNLOCK pthread_mutex_unlock(&mutex);
#endif
#ifdef NO_THREAD
#define _MANAGED_ARRAY_LOCK
#define _MANAGED_ARRAY_UNLOCK
#endif


/** Template class for managed arrays. I.e. arrays with key ids so that searches
can be done on them. Also the array auto-resizes when it fills up.
*/
template <class Type>
class managedArray_c {
public:

	enum status_t {FAILED, OK};
	const int defaultGrowSize = 20;

	~managedArray_c() {
		if (array)
			delete array;
		if (id)
			delete id;
		if (used)
			delete used;

#ifdef NEXUS_THREAD_SAFE
		nexus_mutex_destroy(&mutex);
#endif
#ifdef PTHREAD_SAFE
		pthread_mutex_destroy(&mutex);
#endif

	}

	/// Manually lock this array object. (not recommended)
	void lock() {
		_MANAGED_ARRAY_LOCK
	}

	/// Manually unlock this array object. (not recommended)
	void unlock() {
		_MANAGED_ARRAY_UNLOCK
	}


	managedArray_c() {
		currentSize = -1;
		size = 0;
		growSize = defaultGrowSize;
		array = NULL;
		id = NULL;
		used = NULL;

#ifdef NEXUS_THREAD_SAFE
		nexus_mutex_init(&mutex,NULL);
#endif
#ifdef PTHREAD_SAFE
		pthread_mutex_init(&mutex,NULL);
#endif

	}

	/// Set the grow size each time array needs to allocate more memory.
	status_t setGrowSize(int sze) {
		if (sze < 1) return FAILED;
		defaultGrowSize = sze;
		return OK;
	}

	/// Set size of the array.
	status_t setSize(int sze = defaultGrowSize) {

		lock();

		// If it has not been initialized yet then do so.
		if (size == 0) {

			currentSize = -1;
			size = 0;

			array = new Type[size];
			if (!array) {

				unlock();

				return FAILED;
			}

			id = new _support_hashWord_t[size];
			if (!id) {
				delete array;

				unlock();

				return FAILED;
			}

			used = new int[size];
			if (!used) {
				delete array;
				delete id;

				unlock();

				return FAILED;
			}

			size = sze;

			unlock();

			clear();

			return OK;
		} else {


			// Resize the existing array.

			if (sze <= currentSize) {

				unlock();

				return FAILED;
			}

			Type *newArray;
			_support_hashWord_t *newId;
			int *newUsed;

			newArray = new Type[sze];

			if (!newArray) {
				
				unlock();

				return FAILED;
			}

			newId = new _support_hashWord_t[sze];
			if (!newId) {

				unlock();

				delete newArray;
				return FAILED;
			}


			newUsed = new int[sze];
			if (!newUsed) {

				unlock();

				delete newArray;
				delete newId;
				return FAILED;
			}

			for(int i =0; i < currentSize+1; i++) {
				newArray[i]= array[i];
				newId[i] = id[i];
				newUsed[i] = used[i];
			}
			delete array;
			delete id;
			delete used;
			array = newArray;
			id = newId;
			used = newUsed;
			size = sze;

			unlock();

			return OK;

		}
	}

	/// Clear out the array.
	void clear() {

		lock();

		for(int i =0; i< size; i++) {
			array[i] = NULL;
			id[i] = 0;
			used[i] = 0;
		}
		currentSize = -1;

		unlock();
	}

	/// Find an item based on a search key id.
	Type find(_support_hashWord_t idSearch) {
		Type item = NULL;

		lock();

		for (int i = 0; i < currentSize+1; i++){
			if ((id[i] == idSearch) && (used[i])) {
				item= array[i];
				break;
			}
		}

		unlock();

		return item;
	}

	/// Return the index number of the search item.
	int findNth(_support_hashWord_t idSearch) {
		int val = -1;

		lock();

		for (int i = 0; i < currentSize+1; i++){
			if ((id[i] == idSearch) && (used[i])){
				val = i;
				break;
			}
		}

		unlock();

		return val;
	}

	/// Add an item to the array.
	status_t add(Type item, _support_hashWord_t  idSearch) {


		lock();

		currentSize++;

		if (currentSize >= size) {

			unlock(); 

			currentSize--;
			setSize(currentSize+growSize);
			currentSize++;

			lock();
		}
		array[currentSize] = item;
		id[currentSize] = idSearch;
		used[currentSize] = 1;


		unlock();

		return OK;
	}

	/// Remove an item.
	status_t remove(_support_hashWord_t idSearch) {

		lock();

		int i = findNthNoLock(idSearch);

		if (i == -1) {

			unlock();

			return FAILED;
		}


		if (currentSize == 0){
			used[0] = 0;
		}
		else {
			array[i] = array[currentSize];
			id[i] = id[currentSize];
			used[i] = 1;
			used[currentSize] = 0;
		}
		currentSize--;


		unlock();

		return OK;

	}

	/// Return number of items in the array.
	int numItems() {

		lock();

		int sz = currentSize+1;

		unlock();

		return sz;
	}

	/// Return the nth item in the array.
	Type getNthItem(int i) {return array[i];}

	/// Return the nth id in the array.
	_support_hashWord_t getNthID(int i) {return id[i];}


private:
	Type *array;
	_support_hashWord_t *id;
	int size;
	int currentSize;
	int *used;

	int growSize;

#ifdef NEXUS_THREAD_SAFE
	nexus_mutex_t mutex;
#endif
#ifdef PTHREAD_SAFE
	pthread_mutex_t mutex;
#endif

	// Return the index # of the search item.
	// No mutex locking is done.
	int findNthNoLock(_support_hashWord_t idSearch) {
		int val = -1;
		for (int i = 0; i < currentSize+1; i++){
			if ((id[i] == idSearch) && (used[i])){
				val = i;
				break;
			}
		}
		return val;
	}


};



#endif
