/*
CAVERN API Specification
Draft Version: 4/23/97
Author: Jason Leigh

This is removes the notion of a subscription from the previous 4/22/97 API spec.
Essentially you create a personal IRB.
Then with the personal IRB you can define remote and local keys.
Once you have the keys you can link between them to share data.
The link specifies the networking type of service to use.

Features Missing:

Directory/Hierarchy Operations

Recording State persistence Operations

CAVERN_keyDefAttrib_c to allow the specification of database committment
frequency and/or conditions. 
Permissions and Security

In irb class need calls to change permissions.

read/write locks

policies for shut down and broken connection

in link class need api calls for getting info about the keys.

*/


#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>


class CAVERN_irbKey_c;
class CAVERN_irbId_c;
class CAVERN_irbKeyId_c;
class CAVERN_irbKeyInfo_c;
class CAVERN_irbInfo_c;
class CAVERN_keyDefAttrib_c;
class CAVERN_TCPKey_c;
class CAVERN_UDPKey_c;
class CAVERN_MCASTKey_c;
class CAVERN_WWWKey_c;
class CAVERN_linkAttrib_c;
class CAVERN_key_c;
class CAVERN_keyInfo_c;
class CAVERN_discreteQueue_c;
class CAVERN_stream_c;

/** Default port of all IRBs.
Can be overriden using CAVERN_initAttrib_c.
*/
const unsigned short CAVERN_IRB_DEFAULT_MAIN_PORT = 9821;

/** Class to set CAVERN personal IRB initialization attributes.
With this you can choose the port to open for your personal IRB and you
can choose the max number of threads an IRB may spawn. Other attributes
will be added with time.

*/
class CAVERN_initAttrib_c {
public:
	/// Have CAVERN decide how many threads to spawn.
	const int CAVERN_AUTO_THREAD_MANAGE = 0;

	/** Change CAVERN default main port to something else. Must be
	a number greater than 6000. The main port is the port that all
	IRBs connect to first before establishing any other communication
	lines.
	*/
	void setPort(unsigned short thePort);

	/// Get default main port of CAVERN client.
	unsigned short getPort();

	/** Choose max number of threads CAVERN is allowed to spawn.
	 @param max must be either set to CAVERN_AUTO_THREAD_MANAGE or
	 some number greater than 1.
	 */
	void setMaxThreads(int max);

	/** Get max num threads CAVERN has been allowed to spawn.
	If value is 0 then it means CAVERN is managing this automatically.
	*/
	int getMaxThreads();

	CAVERN_initAttrib_c() {
		maximumNumberOfThreads = CAVERN_AUTO_THREAD_MANAGE;
		mainPort = CAVERN_IRB_DEFAULT_MAIN_PORT;
	}
private:
	int maximumNumberOfThreads;
	unsigned short mainPort;

};


/** Class for personal IRB returned from CAVERNInit().
CAVERNInit() is the only function that generates this object.
Each client or server program only has 1 of these objects.

Once you have this IRB object you can call the CAVERN_irb_c::define()
member function to define local and remote keys in which data may be stored.

You can also link (CAVERN_irbKey_c::link) a local key to a remote key
so that any updates will be shared between the keys.

*/
class CAVERN_irb_c {
public:

	/// Status codes returned from various methods.
	enum statusCodes_t  {
		/// Operation successful.
		OK = -1,

		/// Key does not exist.
		KEY_NOT_EXIST = -2,

		/// Cannot connect to remote IRB.
		CANNOT_CONNECT_TO_REMOTE_IRB = -3,

		/// Permission denied  during subscription.
		PERMISSION_DENIED = -4,

		/// Key exists.
		EXISTS = -5,

		/// Key does not exist.
		NOT_EXIST = -6,

	};

	/** Trigger events for IRBs.
	  @see trigger
	  */
	enum CAVERN_irbTriggerEvent_t  {
		/// Event for when a connection is broken.
		CONNECTION_BROKEN,

		/// Remote IRB has cleanly shutdown.
		SHUTDOWN
	};


	/** Request information about a remote IRB.
	  Note: this information is provided as a snapshot. You need to
	  call this again if you want to monitor it change. 

	  @param remoteIRB address of remote irb. If parameter set to
	  NULL it assumes you mean current personal IRB.

	  @return an object that contains the info if any else returns NULL.
	  Delete the  object when you are done using it.

	  @param status returns status of call which can be:
		OK, CANNOT_CONNECT_TO_REMOTE_IRB, PERMISSiON_DENIED
	  */
	CAVERN_irbInfo_c *requestIRBInfo(const CAVERN_irbId_c *remoteIRB,
				       statusCodes_t *status);

	/** Request information about a key.
	  @param irb is a remote IRB. Set irb to NULL to indicate local personal IRB.

	  @param key is a key (remote or local).

	  Note: this information is provided as a snapshot. You need to
	  call this again if you want to monitor it change.

	  @return an object that contains meta information about a key.
	  Delete this object when you are done using it.
	 */
	CAVERN_irbKeyInfo_c *requestKeyInfo(const CAVERN_irbId_c *irb,
					  const CAVERN_irbKeyId_c *key);


	/** Define a key on a remote IRB.  By defining a remote key
	  that key is declared PERSISTENT.  That means it will
	  continue to exist even when the IRB is stopped and
	  restarted. The only way to remove the key is to undefine it.
	  This simplifies issues on crash recovery.

	 This function does not return a key as you might expect
	 because you must link (see CAVERN_irbKey_c::link) to it
	 before you can access data from it. To do this you need to define a
	 local key so that the remote key and the local can be linked.
	 To define a local key see the other overloading of this define() function.

	 If you define a key that already exists and the attributes
	 of your new definition are different then you will change
	 those attributes on the remote site providing you have the permissions.

	 Defining a key that already exists does NOT clear out its cached contents.

	 @param keyDef set this to NULL for default settings.
	 This is intended to set characteristics such as permissions and
	 frequency and conditions of persistent update.

	 @param status Returns one of the following status codes: OK,
	 CANNOT_CONNECT_TO_REMOTE_IRB, EXISTS, PERMISSION_DENIED
	 */
	void define(const CAVERN_irbId_c *remoteIRB,
		    const CAVERN_irbKeyId_c *newKeyInfo,
		    const CAVERN_keyDefAttrib_c *keyDef,
		    statusCodes_t *status);

	/** Define a key on the personal IRB.

	  If local key already exists and the attributes have changed this
	  call to define simply alters the keys attributes.

	  If local key already exists and the attributes have not changed
	  then this is simply a "reactivation" of the key.

	  @param keyDef if set to NULL then it assumes default settings.
	  If a key already exists then it assumes the setting of the existing
	  key.

	  @return A handle to the database. Delete the handle when you're
	  done with it. Note deleting a handle does not undefine the key
	  in the database. This must be done with the undefine member function.
	  */
	CAVERN_irbKey_c *define(const CAVERN_irbKeyId_c *newKeyInfo,
		    const CAVERN_keyDefAttrib_c *keyDef,
		    statusCodes_t *status);

	/** Undefine a key on an IRB.
	  This only occurs when there are  no longer keys linked to it.
	  If there are it simply fails.
	 */
	void undefine(const CAVERN_irbId_c *remoteIRB,
		     const CAVERN_irbKeyId_c *keyInfo, statusCodes_t *status);

	/// Undefine a local key.
	void undefine(CAVERN_irbId_c *keyInfo, statusCodes_t *status);
	
	/** Determines if a key exists.
	 */
	statusCodes_t exists(const CAVERN_irbId_c *remoteIRB, const CAVERN_irbKeyId_c *keyInfo);

	/** Create a key for direct access to TCP. The personal
	  IRB generates this so that it can keep a tab on it to automate
	  some services. Deleting the object means you are done with the TCP
	  connection.
	  */
	CAVERN_TCPKey_c *makeTCPKey();

	/** Manage a key for direct access to UDP. Deleting the key
	  means you no longer want CAVERN to manage it.
	  */
	CAVERN_UDPKey_c *makeUDPKey();

	/** Manage a key for direct access to MCAST. Deleting the key
	  means you no longer want CAVERN to manage it.
	  */
	CAVERN_MCASTKey_c *makeMCASTKey();

	/** Manage a key for direct access to WEB server using HTTP1.0.
	  Deleting the key  means you no longer want CAVERN to manage it.
	  */
	CAVERN_WWWKey_c *makeWWWKey();

	// Set Global permissions of other IRBs connecting to you.
	//void setPermissions(CAVERN_permissions_t permissions);

	/** Trigger on events.

	  The policies for shutdown and connecion_broken are still
	  being designed.

	  CONNECTION_BROKEN:
	  With this info the user can do whatever is necessary in preparation.
	  By default personal IRB will intermittently try to re-establish
	  connection. There are additional accessor functions that would
	  be helpful here like: Get me a list of keys that are affected by
	  this break in connection. Given the keys you might be able
	  to recover by linking to some other IRB.

	  SHUTDOWN:
	  Clean shutdown. Make it an option to re-establish a connection.

	  */
	void trigger(void (*callback)(CAVERN_irbTriggerEvent_t event,
				      const CAVERN_irbId_c *,
				      void* userData),
		     void* userData);

	/// Shutdown personal IRB.
	void shutdown();
};


/* TODO Need to add something to allow setting permissions for specific IRBs.
I.e. a list of who is allowed to read/write/connect
enum CAVERN_permissions_t  {CAVERN_READ, CAVERN_WRITE, CAVERN_READ_WRITE};
*/

/** Class describing the properties of a key definition.
Used in calling: CAVERN_irb_c::define().

Currently there are no attributes.

Permission-type attributes and frequency and conditions of database
committments will be introduced later.

*/
class CAVERN_keyDefAttrib_c {

public:

private:
	/* TODO
	//More elaborate permission schemes can be added later to the class.
	CAVERN_permissions_t creatorPermissions, otherPermissions;
	*/

	/* OMITTED
	   Set retention type. Retention determines if a key is transient or
	   persistent. If transient then when all subscribers 

	   void setRetention(CAVERN_retention_t reten);
	   CAVERN_retention_t getRetention();
	*/
};

/** Class to define network Quality of Service parameters. QoS attributes allow
you to negotiate how much bandwidth, latency, and jitter is acceptable over
your subscription to a remote IRB.
*/
class CAVERN_qosAttrib_c {

public:
	/** Strategies for acquiring QOS.
	  @see setQosGetRule, getQosGetRule
	  */
	enum CAVERN_qosGetRule_t  {
		/// Get exact QoS desired or fail.
		QOS_GET_EXACT,
		/// Ask for QoS but negotiate for best that can be gotten.
		QOS_NEGOTIATE
	};

	/// Set QoS rule.
	void setQosGetRule(CAVERN_qosGetRule_t);

	/// Set acceptable latency in milliseconds.
	void setAcceptableLatency(float);

	/// Set acceptable jitter in milliseconds.
	void setAcceptableJitter(float);

	/// Set bandwidth in bits per second.
	void setBandwidth(int);

	/// Accessor for QoS rule.
	CAVERN_qosGetRule_t getQosGetRule();

	/// Accessor for latency.
	float getAcceptableLatency();

	/// Accessor for jitter.
	float getAcceptableJitter();

	/// Accessor for bandwidth (in bits per second).
	int getBandwidth();

private:

	CAVERN_qosGetRule_t qosGetRule;
	int bandwidthAllocation;
	float acceptableLatency;
	float acceptableJitter;

};

/** This lists all the keys in the IRB as well as info relevant to the IRB.
Such as: num of threads, port numbers, number of links to keys etc.
This will be added with time.
*/
class CAVERN_irbInfo_c {

public:
	/// Number of keys in IRB.
	int getNumberOfKeys();

	/// Returns an array of key objects.
	CAVERN_irbKeyInfo_c **getListOfKeys();

private:
	int numberOfKeys;
	CAVERN_irbKeyInfo_c **keys;
	/* Delete must make sure to delete each array element */

	/* These will be accessible by friend classes who will populate
	   the data.
	   */
	void setNumberOfKeys(int);
	void setListOfKeys(CAVERN_irbKeyInfo_c *);

};

/** Class to specify a key. */
class CAVERN_irbKeyId_c {

public:

	/// Get the comment field describing the key.
	char *getComment();

	/// Get the designated path of key.
	char *getPath();

	/// Get text name of key.
	char *getName();

	/// Set the comment field describing the key.
	void setComment(char*);

	/// Set the designated path of key.
	void setPath(char*);

	/// Set text name of key.
	void setName(char*);

private:
	char *comment;
	// more needed here. i.e. char key string, hash value

	char *path; 
	// This provides some way in the future to generate
	// some kind of hierarchy in the data
	// like paths in a URL or Windows registry.

	char *name;

};


/** Meta info about a key. More fields will be added over time as user's ask
for them.
*/
class CAVERN_irbKeyInfo_c {
public:

	/// Get key id describing the key. This allocates an object which
	/// you should delete when done.
	CAVERN_irbKeyId_c *getKeyId();

	/// Display parameters of key to stdout.
	void display();

	/// Possible types of keys.
	enum CAVERN_keyType_t  {
		/// The basic key to an IRB. @see CAVERN_irbKey_c
		IRB_KEY,
		/// Key for WEB access. @see CAVERN_WWWKey_c
		WWW_KEY,
		/// Key for direct connections over TCP/IP. @see CAVERN_TCPKey_c
		TCP_KEY,
		/// Key for direct connections over UDP. @see CAVERN_UDPKey_c
		UDP_KEY,
		/// Key for direct connections over multicast. @see CAVERN_MCASTKey_c
		MCAST_KEY
	};

	/// Figure out what kind of key do we have.
	CAVERN_keyType_t getKeyType();

	/** Get the subscription if any, else return NULL.
	  DO NOT ALLOW THIS. If they want a handle on the key they need to
	  use define or makeTCP etc... This getKey will screw a lot of things
	  up.
	 Returns a key either of type: CAVERN_irbKey_c,
	 CAVERN_WWWKey_c, CAVERN_TCPKey_c, CAVERN_UDPKey_c, CAVERN_MCASTKey_c

	CAVERN_key_c *getKey();
	*/

	/// Get size of item currently stored in key.
	int getSize();

	/// Get time stamp of current key.
	double getTimeStamp();

	/// Get total size of incoming replacement for key.
	int getIncomingTotalSize();

	/// Get size downloaded so far of incoming replacement for key.
	int getIncomingCurrentSize();

	/// Get time stamp of incoming replacement for key.
	double getIncomingTimeStamp();

	/// Get current QoS attributes.
	CAVERN_qosAttrib_c *getQos();

	/// Get contracted QoS attributes.
	CAVERN_qosAttrib_c getContractedQos();

	/* OBSOLETE: Returns a handle to the subscription if exists, else returns NULL.
	   DISALLOW THIS. Since handles are now envelopes this would generate
	   another envelope. Also it is stupid to implement
	   semantics where it returns something if it has already been
	   subscribed to else NULL since by common sense you'd expect
	   it to give you the subscription. But you can't just be given
	   the subscription because there are attributes that need to be specified
	   with a subscribe() call.
	CAVERN_key_c * getSubscription();
	*/

	/** Returns the number of links connected to this key from remote
	clients.
	*/
	int getNumberOfLinks();

private:
	CAVERN_irbKeyId_c keyId;
	int keyHashID;
	int size;
	int incomingSize;
	int incomingCurrentSize;
	int incomingTotalSize;
	double timeStamp;
	CAVERN_qosAttrib_c currentQos;
	CAVERN_qosAttrib_c contractedQos;

	/*
	// Need to make these friends of the class that populates it
	// with data. That is to be determined.
	void setSize(int);
	void setTimeStamp(double);
	void setIncomingTotalSize(int);
	void setIncomingCurrentSize(int);
	void setIncomingTimeStamp(double);
	*/

	// This is either filled with a pointer to a local subscription or NULL.
	// CAVERN_key_c *theSubscriptionIfAny;

	int numberOfLinksToKey;

};

/** Base class for CAVERN keys. Do not instantiate any objects of this type */
class CAVERN_key_c {

};

/** CAVERN IRB Key class. Objects of this type are instantiated by
calling CAVERN_irb_c::define(). Deletion of this object does not
mean the key is undefined. It simply means you have deleted the handle.
You can reaquire the handle by calling CAVERN_irb_c::define().

Using a key you may link with remote keys so that data can be shared
transparently between them. See CAVERN_irbKey_c::link().
*/
class CAVERN_irbKey_c : public CAVERN_key_c {

public:

	/// Status codes returned from various methods.
	enum statusCodes_t  {
		/// Operation succeeded.
		OK,
		/// Operation failed.
		FAILED
	};

	/// Blocking/Non-blocking attribute type.
	enum CAVERN_blocking_t  {
		/// Make an operation blocking.
		BLOCKING,
		/// Make an operation non-blocking (ie return immediately)
		NON_BLOCKING
	};

	/// Trigger event types for a key.
	enum CAVERN_keyTriggerEvent_t  {

		/// Triggers just before newly arrived data replaces current data in key.
		PRE_NEW_DATA_ARRIVED,

		/// Triggers after newly arrived data has replaced data in key.
		POST_NEW_DATA_ARRIVED,

		/** This says that remote data is availabe but is not
		  downloaded.  For passive updates the data is not
		  sent but a notification of a newer version of the
		  data is available is sent.
		  */
		NEW_DATA_REMOTELY_AVAILABLE

	};

	/** Link remote key to local key.
	  @param linkAttr specifies the attributes for the link.
	  It specifies active/passive update, network QoS, network reliability.
	  @return Returns a link object. You must delete it when you
	  are done with it. See CAVERN_irbLink_c for more details.
	  */
	CAVERN_irbLink_c *link(const CAVERN_irbId_c *remoteIRB,
				    const CAVERN_irbKeyId_c *remoteKey,
				    CAVERN_linkAttrib_c *optionalLinkAttr,
				    statusCodes_t *status);

	/** Get data from key.
	  @param sze passed as pointer. Returns sze of data returned.
	  If key has no data yet, sze returns 0.
	  @param blockingAttrb determines if the get operation will block
	  if no data is in the key. Default is blocking.
	  @return data buffer of key. It is the user's responsibility to
	  deallocate the memory when done. Returns NULL if key has no data yet.
	  */

	char* get(int *sze, CAVERN_blocking_t blockingAttrib = BLOCKING);

	/** Request the data from a remote source if there is a link with a remote source.
	  Used for passive updates. This call will be ignored for active updates.
	 */
	void request(CAVERN_blocking_t blockingAttrib);

	/** Get subpart of data from key.
	  Note, this usually implies the data is rather large so some form of
	  a "sparse virtual memory emulation" will be used to "page-in" segments of
	  the imported data.
	  @param startIndex beginning of subsection to extract.
	  @param endIndex end of subsection to extract. If endIndex is NULL it means
	  extract to the end. endIndex must be either NULL or greater than or equal to
	  startIndex.
	  @param sze passed as pointer. Returns sze of data extracted.
	  If key has no data yet returns 0.
	  @return data buffer of subsection. It is the user's responsibility to
	  deallocate the memory when done.
	  */
	char* get(int startIndex, int endIndex, int *sze,
		  CAVERN_blocking_t blockingAttrib);

	/// Request subpart of data from remote source.
	void request(int startIndex, int endIndex,
		     CAVERN_blocking_t blockingAttrib);

	/** Put data into key. This replaces the previous contents with the
	  new contents. This call usually results in a copy of the entire
	  data to the remote site.
	  @param buffer gets copied to key.
	  @param sze amount of data to copy. Returns sze successfully copied.
	  */
	void put(char *buffer, int *sze);

	/** Put data into subpart of key. This usually results in the
	  transmission of the subpart rather than the whole piece of data since
	  this is likely to be used to access a large data set.
	  If Put is done when there is incoming data then the put
	  places the data in the local key, and ignores the incoming
	  data stream if the local timestamp is more current.
	  @param buffer gets copied to key.
	  @param startIndex is location in key to begin copying.
	  @param sze amount of data to copy. Returns sze successfully copied.
	  If sze exceeds current sze of data in key, data cell is automatically expanded.
	  */
	void put(char *buffer, int startIndex, int *sze);

	/** Append data to data in key. Only the newly appended extra data
	  is transmitted. Note when you use calls like Append and Put
	  you run the risk of inconsistencies when there are multiple clients
	  updating the same information. It is advisable to use locks in those
	  situations.
	  @param buffer copied and appended to data in key.
	  @param sze amount of data to append. Returns sze successfully appended.
	  */
	void append(char *buffer, int *sze);

	/** Deliver data inside current key.
	  This is for passive updates over a link. Links with active updates
	  ignore this call.
	  */
	void deliver();

	/** Get info about this key.
	  @param keyInfo will return information about the state of the key.
	  @see CAVERN_keyInfo_c for an explanation of status.
	  */
	CAVERN_irbKeyInfo_c getInfo();

	/// Look at size of data in key.  getInfo will do this too but with more detail.
	int getSize();

	/** Trigger on an event.
	  @param callback is the callback to call when trigger occurs.
	  Callback is of the form:
	  void (*callback)(CAVERN_triggerEvent_t event, CAVERN_key_c *thisKey,
						void* userData);
	  @param userData is address of data to pass into callback.
	  */
	void trigger(void (*callback)(CAVERN_keyTriggerEvent_t event,
				      CAVERN_irbKey_c *thisKey,
				      void* userData),
		     void* userData);

	/** Lock this key. 
	  The lock works as follows: First this call blocks while it checks the lock.
	  If it is not locked then it grabs the lock.
	  If the item is already locked it
	  can be set to block further until lock becomes available or return
	  with a failure to lock.

	  Note: Eventually we will also provide read/write locks to improve
	  concurrency.

	  @return The return value is status of the lock attempt: either OK or FAILED
	  */
	statusCodes_t lock(CAVERN_blocking_t blockingParam);

	/// Lock the remote key that is linked to this key (if any).
	statusCodes_t lockRemote(CAVERN_blocking_t blockingParam);

	/** Unlock a previously acquired lock.
	  @see lock().
	  */
	void unlock();

	/// UnLock the remote key that is linked to this key.
	void unlockRemote();

	/** Tell IRB to manage the queue using data from this key.
	  WARNING: For large keys this can eat up virtual memory fast!
	  Deleting the queue means you no longer want the IRB to manage
	  it.
	  */
	CAVERN_discreteQueue_c * queue(statusCodes_t status);

	/** Tell IRB to manage stream using data from this key.
	  WARNING: For large keys this can eat up virtual memory fast!
	  Deleting the stream means you no longer want the IRB to manage it.
	 */
	CAVERN_stream_c *queue(statusCodes_t status);

};

/** Class to specify the characteristics of a link between two keys.
Characteristics that can be specified include:
network reliability, and active/passive updating.
QoS is not defined here. It is set in CAVERN_irbLink_c since it can involve
a process of negotiation.
*/
class CAVERN_linkAttrib_c {
 public:
	/// Network reliability types
	enum CAVERN_networkReliability_t  {
		/// Use reliable protocols- DEFAULT.
		RELIABLE,
		/// Use unreliable protocols.
		UNRELIABLE
	};

	/// Update types
	enum CAVERN_update_t  {
		/// Active update- DEFAULT.
		ACTIVE_UPDATE,
		/// Passive update.
		PASSIVE_UPDATE
	};


	/** Set update type of link. Either active or passive update.
	  For active updates any changes in the data will automatically
	  be forwarded to the end point of the link. Also any change at the
	  end point will be returned.

	  For passive updates only
	  an indication that the current linked data is obsolete is
	  forwarded. It is recommended that for small pieces of data
	  active updates are used and for large data passive updates.
	  */
	void setUpdateType(CAVERN_update_t);


	/** Set the delivery reliability to use. Either choose reliable or
	  unreliable. Reliable is slower but guarantees data arrival.
	  Unreliable is faster but does not guarantee data arrival nor
	  does it guarantee order. It does guarantee that data that arrives,
	  if it arrives is correct data.

	  CAVERN detemines what is the best medium for transmission
	  based on networking capability and available resources.  So
	  e.g. if you choose reliable it will try reliable multicast
	  if there are more than 2 users on the system. Else it
	  chooses unicast. If multicast is unavailable it defaults to
	  broadcast by unicasting to all subscribers.  For unreliable
	  connections a similar approach is made first with unreliable
	  multicast, then with broadcasting via udp.  Note: A
	  multicast or unicast channel may carry data for more than one
	  key.
	  */
	void setNetworkReliability(CAVERN_networkReliability_t);

	/// Get network reliability attribute.
	CAVERN_networkReliability_t getNetworkReliability();


};

class CAVERN_irbLink_c {
public:
	/// Status codes returned from various methods.
	enum statusCodes_t  {
		/// Operation succeeded.
		OK,
		/// Operation failed.
		FAILED,
		/// Used for QoS- operation negotiated a lower QoS.
		NEGOTIATED
	};


	/// Trigger events
	enum CAVERN_linkTriggerEvent_t  {
		/// Broken link due to break of connection.
		BROKEN_LINK,

		/// Triggers when the QoS of the network service begins to deviate.
		QOS_DEVIATION
	};

	/// Deleting this link will unlink between the local and remote keys.
	~CAVERN_irbLink_c();

	/** Researve Quality of Service
	@param CAVERN_qosAttrib_c Specifies the desired
	QoS. It returns set with the QoS it was able to negotiate.
	@return OK if got contract ; NEGOTIATED if had to negotiate for lower ;
	FAILED if completely failed.
	*/
	statusCodes_t  researveQoS(CAVERN_qosAttrib_c *);

	/** Suspend active updates. You can hence use this as a form of flow control
	for incoming data.
	@see resumeActiveUpdates
	*/
	void suspendActiveUpdates();

	/** Resume active updates.
	  @see suspendActiveUpdates
	  */
	void resumeActiveUpdates();

	/** Trigger on events about the link.
	  For example, if a connection breaks the user can be warned.
	  The system will attempt to re-establish the link, but you
	  can override this by first unlinking and then forming a new
	  link with perhaps another IRB.

	Need API calls to get info on the linked keys.
	*/
	void trigger(void (*callback)(CAVERN_linkTriggerEvent_t event,
				      CAVERN_irbLink_c *thisLink,
				      void* userData),
		     void* userData);

	/// Get a handle on the local key bound to this link.
	CAVERN_irbKey_c *getLocalKey();

	/// Get an id on the remote IRB bound to this link.
	CAVERN_irbId_c *getRemoteIRBID();

	/// Get an id on the remote key bound to this link.
	CAVERN_irbKeyId_c *getRemoteKeyID();
};


/** CAVERN Queue Base Class.
Do not allocate an object of this type */
class CAVERN_queue_c {

public:

	/* IGNORE THIS Queue items arriving at the key. 
	  @param keyToQueue This only binds to proper CAVERN database keys.
	  It does not allow binding to direct TCP, UDP, MCAST keys.
	  Also does not allow binding to keys subscribed as
	  UPDATE_BY_PARTIAL_REPLICATION
	virtual statusCodes_t Bind(CAVERN_irbKey_c keyToQueue) = 0;
	  */

	/// Clear the queue.
	virtual void clear() = 0;

	/// Peek at the length of queue.
	virtual int peekLength() = 0;

	/** Determine if a queue is empty.
	@return 1 if empty, else returns 0.
	*/
	virtual int isEmpty() = 0;

};

/** CAVERN discrete queue class.
This takes successive discrete data values entering
a key and queues them.

 */
class CAVERN_discreteQueue_c : public CAVERN_queue_c {

public:
	/** Get data from front of queue.
	  @param sze passed as pointer. Returns sze of data in dequeued object.
	  If queue is empty sze returns 0.
	  @return data buffer of queued object. It is the user's responsibility to
	  deallocate the memory when done. Returns NULL if queue is empty.
	  */
	char* get(int *sze);

	/** Get subpart of data from front of queue. The queued item is then completely
	  dequeued.
	  @param startIndex beginning of subsection to extract.
	  @param endIndex end of subsection to extract. If endIndex is NULL it means
	  extract to the end. endIndex must be either NULL or greater than or equal to
	  startIndex.
	  @param sze passed as pointer. Returns sze of data extracted.
	  If queue is empty sze returns 0.
	  @return data buffer of subsection. It is the user's responsibility to
	  deallocate the memory when done.
	  */
	char* get(int startIndex, int endIndex, int *sze);

	/** Peek at queued item without dequeueing it.
	  @see get(int *sze)
	  */
	char* peek(int *sze);

	/** Peek at subpart of queued item without dequeueing it.
	 @see get(int startIndex, int endIndex, int *sze)
	 */
	char* peek(int startIndex, int endIndex, int *sze);

	/// Dequeue an item.
	void deQueue();

	/// Peek at the sze of item at front of queue.
	int peekQueueItemSize();

	/// Clear the queue.
	void clear();

	/// Peek at the length of queue.
	int peekLength();

	/** Determine if a queue is empty.
	@return 1 if empty, else returns 0.
	*/
	int isEmpty();
	
};

/** CAVERN stream class. This takes successive discrete data values entering
a key and queues them so that it can be accessed as a stream.
*/
class CAVERN_stream_c : public CAVERN_queue_c {

public:
	/** Get data from front of queue.
	  @param sze passed as pointer. Pass in the number of bytes you wish to
	  extract. Returns sze of data extracted.
	  If queue is empty sze returns 0.
	  @return data buffer of queued chunk. It is the user's responsibility to
	  deallocate the memory when done. Returns NULL if queue is empty.
	  */
	char* get(int *sze);

	/// Clear the queue.
	virtual void clear();

	/// Peek at the length of queue.
	virtual int peekLength();

	/** Determine if a queue is empty.
	@return 1 if empty, else returns 0.
	*/
	virtual int isEmpty();
	

};


/** Key to support download of WEB documents. By simply filling in the WEB site's
address, path and filename, the IRB will download the data and store it in
the key for your retrieval. Timestamps are also compared to reduce the need to
download redundant data.
 */
class CAVERN_WWWKey_c : public CAVERN_key_c {
public:

	/// Trigger events for WWW key.
	enum CAVERN_triggerWWWEvent_t  {

		/// WEB access timed out but a cached version exists.
		EVENT_TIME_OUT_BUT_CACHE_EXISTS = -1,

		/// WEB timed out but cached version does NOT exist.
		EVENT_TIME_OUT_BUT_CACHE_NOT_EXIST = -2,

		/// Cannot connect to WEB server.
		EVENT_CANNOT_CONNECT = -3,

		/// Using CHECK_BUT_DONT_DOWNLOAD this tells you that the remote version is newer.
		EVENT_REMOTE_VERSION_NEWER = -4,

		/// No download is needed becoz cached version is up-to-date.
		EVENT_NO_DOWNLOAD_NEEDED = -5,

		/** This means your download attempt failed. Your WEB server
		is sending back non HTTP1.0 compliant data.
		*/
		EVENT_UNABLE_TO_GET_LENGTH_OF_WEB_FILE = -6,

		/// Remote file not found.
		EVENT_REMOTE_FILE_NOT_FOUND = -7,

		/// Downloaded successfully.
		EVENT_DOWNLOADED = -8,

		/// For nonblocking calls to Request. @see request.
		EVENT_PROCESSING = -9,
	};

	
	/// Blocking/Non-blocking attribute type.
	enum CAVERN_blocking_t  {
		/// Make an operation blocking.
		BLOCKING,
		/// Make an operation non-blocking (ie return immediately)
		NON_BLOCKING
	};

	/// Used in setTimeOut() to turn off timeouts during WEB download.
	const int NEVER_TIME_OUT = -1;

	/// Determines how the download should proceed. @see setMode.
	enum downloadMode_t  {
		/// Download by comparing timestamps.
		AUTO_LOAD,
		/// Download always regardless of timestamp.
		FORCED_LOAD,
		/// Check the timestamps but don't download.
		CHECK_BUT_DONT_LOAD
	};

	/// Set remote site of download.
	void setRemoteSite(char*);
	/// Accessor for remote site.
	char* getRemoteSite();

	/// Set remote path of download.
	void setRemotePath(char*);

	//// Accessor for remote path.
	char *getRemotePath();

	/// Set remote file to download.
	void setRemoteFile(char*);

	/// Accessor for remote file.
	char *getRemoteFile();

	/** Set timeout period for download of WEB file. Set to NEVER_TIME_OUT
	if no time out is desired.
	*/
	void setTimeOut(int timeout);

	/// Accessor for timeout.
	int getTimeOut();

	/// Set the mode for downloading WEB file.
	void setMode(downloadMode_t mode);

	/// Accessor for mode.
	downloadMode_t getMode();

	/** Request WEB file. If non blocking then this call returns
	  PROCESSING.
	  If it is blocking then it will block until data is gotten or
	  it has failed.
	  */
	CAVERN_triggerWWWEvent_t request(CAVERN_blocking_t blockingType);

	/** Trigger on any events generated by this key.
	  The callback is called with the status code that caused the trigger.
	  */
	void trigger(void (*callback)(CAVERN_triggerWWWEvent_t event,
				      CAVERN_WWWKey_c *thisKey,
				      void* userData),
		     void* userData);

	/// Get downloaded data.
	char *get(int *sze);
};


/** CAVERN direct support for TCP/IP. It is not recommended that you use this
except to connect to clients and servers that do not use IRBs (legacy systems).
TCP is a reliable streaming protocol.

This object is instantiated by calling: CAVERN_irb_c::makeTCPKey().
Once you have this key you can choose to make it a client connection or a server.

To make it a client, call the connect()  member function. Then you may call
put or get as needed.

To make it a server, call the accept() member function. Each time a remote
client connects to you a user-defined callback will be fired and a new TCP
key is allocated. You can then use the new key to communicate with the connection.
You should not do calls to get and put with the server key. You should only
do it with the keys generated from the accept() callback.

Finally by using a trigger you can be asynchronously notified of when a key
has new incoming data.


*/
class CAVERN_TCPKey_c : public CAVERN_key_c {

public:

	/// Events that might occur in TCP connections.
	enum CAVERN_TCP_triggerEvent_t  {

		/// New data has arrived.
		NEW_DATA_ARRIVED,

		/// A connection has closed.
		CONNECTION_CLOSED_EVENT
	};

	/// Status codes.
	enum statusCodes_t  {

		/// Operation successful.
		OK = 0,
		/// Operation failed.
		FAILED = -1,
		/// QoS negotiation made.
		NEGOTIATED = -2,

		/** For nonblocking socket calls this is the status if
		a read is attempted on a socket with no data available to be read.
		*/
		NON_BLOCKING_NO_DATA_TO_READ = -3,

		/// Connection closed during get/put call.
		CONNECTION_CLOSED = -4,

		/// Reading or writing data timedout.
		TIME_OUT = -5
	};

	/// Used in setTimeOut() to turn off timeouts during WEB download.
	const int NEVER_TIME_OUT = -1;

	/// Blocking/Non-blocking attribute type.
	enum CAVERN_blocking_t  {
		/// Make an operation blocking.
		BLOCKING,
		/// Make an operation non-blocking (ie return immediately)
		NON_BLOCKING
	};

	/** If connect is called then assuming role of a client connection.
	  and hence connect to a remote server. An accept call assumes you are
	  taking the role of a server. You can not perform a connect() and
	  an accept() together on a single key.
	  */
	statusCodes_t connect(char *ip, unsigned short port);

	/** Researve Quality of Service
	@param Desired qos attributes. Also returns with contract.
	@return OK if got contract ; NEGOTIATED if had to negotiate for lower ;
	FAILED if completely failed.
	*/
	statusCodes_t reserveQoS(CAVERN_qosAttrib_c *qos);

	/** Accept incoming TCP connections.
	  If accept is called then assuming role of a server.

	  @param callback gets called if a new incoming connection is established.

	  @param newConnectionKey is the key that points to the newly
	  established connection.  newConnectionKey is the dynamically
	  generated key for the new connection.

	  Callback's thisKey parameter contains a pointer to the key that
	  accepted the incoming connection- namely this key.

	  @param userData is data that the user may pass when the callback is called.  */
	void accept(unsigned short port,
		   void (*callback)(CAVERN_TCPKey_c *thisKey,
				    CAVERN_TCPKey_c *newConnectionKey, void* userData),
		   void *userData);

	/** Trigger on TCP events.
	 */
	void trigger(void (*callback)(CAVERN_TCP_triggerEvent_t event,
				      CAVERN_TCPKey_c *thisKey,
				      void* userData),
			void* userData);

	/** Get data from socket.
	@return status:
	Returns a status code or else status = UNIX errno.
	@see statusCodes_t

	@param buffer to fill with data.
	@param sizeOfBuffer specifies the number of bytes to put into buffer.
	It also returns the number of bytes successfully extracted.
	@param howto determines if the call blocks until a timeout period or is
	non blocking.
	@param timeOut is timeout period that get blocks waiting for data. If connection
	is non blocking then timeOut is the period it waits once it has started getting
	data. i.e. if a non blocking get is attempted when no data is available, it returns
	immediately and returns a status of -1.
	*/
	statusCodes_t get(char * buffer, int *sizeOfBuffer, CAVERN_blocking_t howto,
		int timeOut = NEVER_TIME_OUT);

	/** Send data over TCP.
	  @param buffer contains data to send.
	  @param sizeOfBuffer contains amount of data in buffer to send. Returns
	  the amount of data successfully sent.
	  @return status:
	  Returns a status code or else status = UNIX errno.
	  @see statusCodes_t
	  */
	statusCodes_t put(char *buffer, int *sizeOfBuffer);

	/// Return IP address of self.
	unsigned long getSelfIP();

	/// Return IP address of self as a string xxx.xxx.xxx.xxx.
	void getSelfIP(char *ip);

	/// Return port of self.
	unsigned short getSelfPort();

	/** Return IP address of remote connection. If this is a server object
	then this returns 0.
	*/
	unsigned long getRemoteIP();

	/// Return IP address of remote connection as a string xxx.xxx.xxx.xxx.
	void getRemoteIP(char *ip);

	/// Return port of remote connection.
	unsigned short getRemotePort();

	/// Disconnect socket connection and tell IRB to unmanage it.
	void disconnect();
};


/** CAVERN direct support for UDP. It is not recommended that you use this
except to connect to clients and servers that do not use IRBs (legacy systems).
UDP is an unreliable packet protocol. Keep the packets small.

You can use this either as a client or a server like in TCP but remember UDP
is connectionless.

To open the connection as a client so that you can send data out to remote sites,
call open(). Then call setSendAddress(). Once this is done you can send a packet to
that address. You can at anytime change the send address by calling setSendAddress().

To open as a server call open(). Usually you want to call open with a specific address
so that the remote clients have an exact port to connect to. Once opened, simply
wait for incoming data. You can use the trigger to notify you when that happens.

For each piece of incoming data you can find out where they came
from. By calling copyReceiveAddressToSendAddress()
you can then send a message back to the original sender.

*/
class CAVERN_UDPKey_c : public CAVERN_key_c {
public:

	/// Status codes for UDP operations.
	enum statusCodes_t  {

		/// Operation succeeded.
		OK = 0,
		/// Operation failed.
		FAILED = -1,
		/// A lower QoS service  was negotiated.
		NEGOTIATED = -2,

		/** For nonblocking socket calls this is the status if
		a read is attempted on a socket with no data available to be read.
		*/
		NON_BLOCKING_NO_DATA_TO_READ = -3
	};

	/// Trigger Events for UDP.
	enum CAVERN_UDP_triggerEvent_t  {

		/// New data has arrived.
		NEW_DATA_ARRIVED
	};

	/** Open UDP socket */
	statusCodes_t open(int port=0); // port = 0 for dynamic port assignment

	/** Choose a destination to send data */
	statusCodes_t setSendAddress(char *ip, int port);

	/** Researve Quality of Service
	@param Desired qos attributes. Also returns with contract.
	@return OK if got contract ; NEGOTIATED if had to negotiate for lower ;
	FAILED if completely failed.
	*/
	statusCodes_t reserveQoS(CAVERN_qosAttrib_c *qos);

	/** Trigger on UDP events. There is currently only 1 event NEW_DATA_ARRIVED.
	  There is no CONNECTION_CLOSED since UDP is connectionless.
	 */
	void trigger(void (*callback)(CAVERN_UDP_triggerEvent_t event,
				      CAVERN_UDPKey_c *thisKey,
				      void* userData),
			void* userData);

	/** Get data.
	  @return status
	  Returns a status code or else status = UNIX errno.
	  @see statusCodes_t

	  @param buffer to fill with data.
	  @param sizeOfBuffer specifies the number of bytes to put into buffer.
	  It also returns the number of bytes successfully extracted.
	  */
	statusCodes_t get(char * buffer, int *sizeOfBuffer);

	/** Send data.
	  @param buffer contains data to send.
	  @param sizeOfBuffer contains amount of data in buffer to send. Returns
	  the amount of data successfully sent.
	  @return status:
	  Returns a status code or else status = UNIX errno.
	  @see statusCodes_t
	  */
	statusCodes_t put(char *buffer, int *sizeOfBuffer);

	/** Take an incoming packet's address and make this the outgoing address
	  so that you can send data directly back to them.
	  An alternative method is to instantiate a new UDP object and set the
	  destination using the setSendAddress call.
	  */
	void copyReceiveAddressToSendAddress();

	/// Get an incoming message's ip and port.
	//@{
	/// Get incoming message's ip.
	unsigned long getReceiveIP();
	/// Get incoming message's ip as a string.
	void getReceiveIP(char *ip);
	/// Get incoming message's port.
	unsigned short getReceivePort();
	//@}


	/// Get destination's ip and port.
	//@{
	/// Get destination ip.
	unsigned long getSendIP();
	/// Get destination ip as string.
	void getSendIP(char *ip);
	/// Get destination port.
	unsigned short getSendPort();
	//@}

	/// Get this key's ip and port.a
	//@{
	/// Get self ip.
	unsigned long getSelfIP();
	/// Get self ip  as string.
	void getSelfIP(char *ip);
	/// Get self port.
	unsigned short getSelfPort();
	//@}

	/// Close udp key and tell IRB to unmanage it.
	void close();
};

/** CAVERN direct support for multicast.
It is not recommended that you use this
except to connect to clients and servers that do not use IRBs (legacy systems).

Using multicast connections are the simplest. Simply open a connection to a specific
multicast group and port number. Multicast group address must begin with 224.xxx.xxx.xxx
and port numbers should be greater than 6000.

Once open, you can send information to all subscribers or get information by
using put and get respectively.

You may also trigger on any incoming data.

Multicast is an connectionless protocol so you won't
know if a connection has been broken. Since it is an unreliable protocol
you should keep the packets small.
*/
class CAVERN_MCASTKey_c : public CAVERN_key_c {
public:
	/// Status codes for UDP operations.
	enum statusCodes_t  {

		/// Operation succeeded.
		OK = 0,
		/// Operation failed.
		FAILED = -1,
		/// A lower QoS service  was negotiated.
		NEGOTIATED = -2,

		/** For nonblocking socket calls this is the status if
		a read is attempted on a socket with no data available to be read.
		*/
		NON_BLOCKING_NO_DATA_TO_READ = -3
	};

	/// Connect to multicast group.
	statusCodes_t open(char *ip, int port);

	/// Trigger Events for UDP.
	enum CAVERN_MCAST_triggerEvent_t  {
		/// New data has arrived.
		NEW_DATA_ARRIVED
	};

	/** Researve Quality of Service
	@param Desired qos attributes. Also returns with contract.
	@return OK if got contract ; NEGOTIATED if had to negotiate for lower ;
	FAILED if completely failed.
	*/
	statusCodes_t reserveQoS(CAVERN_qosAttrib_c *qos);

	/** Trigger on MCAST events. There is currently only 1 event NEW_DATA_ARRIVED.
	  There is no CONNECTION_CLOSED since MCAST is connectionless.
	 */
	void trigger(void (*callback)(CAVERN_MCAST_triggerEvent_t event,
				      CAVERN_MCASTKey_c *thisKey,
				      void* userData),
			void* userData);

	/** Get data.
	  @return status
	  Returns a status code or else status = UNIX errno.
	  @see statusCodes_t

	  @param buffer to fill with data.
	  @param sizeOfBuffer specifies the number of bytes to put into buffer.
	  It also returns the number of bytes successfully extracted.
	  */
	statusCodes_t get(char * buffer, int *sizeOfBuffer);

	/** Send data.
	  @param buffer contains data to send.
	  @param sizeOfBuffer contains amount of data in buffer to send. Returns
	  the amount of data successfully sent.
	  @return status:
	  Returns a status code or else status = UNIX errno.
	  @see statusCodes_t
	  */
	statusCodes_t put(char *buffer, int *sizeOfBuffer);

	/// Get an incoming message's ip and port.
	//@{
	/// Get incoming message's ip.
	unsigned long getReceiveIP();
	/// Get incoming message's ip as a string.
	void getReceiveIP(char *ip);
	/// Get incoming message's port.
	unsigned short getReceivePort();
	//@}


	/// Get mcast destination's ip and port.
	//@{
	/// Get mcast destination ip.
	unsigned long getSendIP();
	/// Get mcast destination ip as string.
	void getSendIP(char *ip);
	/// Get mcast destination port.
	unsigned short getSendPort();
	//@}

	/// Get this key's ip and port.a
	//@{
	/// Get self ip.
	unsigned long getSelfIP();
	/// Get self ip  as string.
	void getSelfIP(char *ip);
	/// Get self port.
	unsigned short getSelfPort();
	//@}

	/// Close mcast key and tell IRB to unmanage it.
	void close();
};

/** Class to specify a remote IRB to connect.
For now this is identified by IP and Port. In the future
we can use Globus to do it.
*/
class CAVERN_irbId_c {

public:

	/// Set address and port of remote IRB.
	void setAddressAndPort(char *ipAddr,
			       unsigned short port = CAVERN_IRB_DEFAULT_MAIN_PORT);

	/// Get address.
	char *getAddress();

	/// Get Port.
	unsigned short getPort();

	CAVERN_irbId_c() {
		port = CAVERN_IRB_DEFAULT_MAIN_PORT;
	}

private:

	char ip[256];
	unsigned short port;

};



/** This is the first thing you do to start CAVERN.
Every CAVERN client has a personal IRB.
This call returns a handle to the personal IRB or NULL if the
initialization failed.

The personal IRB is the main object you use to control CAVERN. See
CAVERN_irb_c for a description of what this handle can do.

WARNING: CAVERN uses threads. Therefore it is important that CAVERNInit()
is called after the LAST fork in your program. For example,
in the CAVE you call CAVERNInit() just before the CAVE's main process'
while loop. This is so that the threads belong to the main
process of the CAVE. What this also means is that CAVERN calls should
ONLY be made in the CAVE main process and NOT the draw processes. But what
this also means is that all variables declared in the main process
are also accessible in the CAVERN thread.

If someday the CAVE library gets rewritten with threads then we will
all be one happy family.

@param attr The attribute sets any special conditions for initializing
CAVERN. If it is set to NULL then default settings are assumed.

*/
CAVERN_irb_c *CAVERNInit(const CAVERN_initAttrib_c *attr = NULL);


/* TODO -------------------------------------------------------------
class CAVERN_mutex_c {
	void lock();
	void unlock();
	int tryLock(); // Same as lock except nonblocking. Returns 1 if locked else 0.
};

*/

//--------------------------------------------------------------------------------------

/* Version 4/15/97 
Ok, this version attempts to remove the need for user to explicitly
connect to other IRBs. Connections are formed dynamically through
Subscriptions.

Also subscriptions determines the type of object acquired.
*/

/*
TODO:
>> New Queue interface
DB hierarchy operations - Hierarchical DB API????
Need IRB calls to do operations on hierarchies. Like get me all keys
under a certain directory etc...

DirectConnect API, WEB API

What about an RPC kind of call. CAVERN is currently data driven, ie
given data at a key do a callback. What about procedure driven, like RCP
and Nexus.

Recording Calls
Threading API, locking etc.. - is there a way to use pthread API. Does Nexus provide one.
Avatar Template API


I wanted the ability for an IRB to subscribe to another IRB's key and if
the remote key is actually a subscription to yet another remote key
it would be nice if there was a way to automatically reroute your subscription
there (if desired). Problem with this is that: as the API currently stands
the pIRB has to call a Connect to remote IRB before it can start subscribing
to data. With the rerouting idea you have to drop the notion of
the connect() call since connects are now made dynamically through
subscription. So in essence you do away with the handle on the remote IRB.

Without the handle to the remote IRB how can you do calls like:
GetMetaInfo on IRB.

NEED TO STRAIGHTEN OUT when data structures get copied and when returned data is copied.
etc....


TODO:


*/


