/* Author: Jason Leigh
 * (C) 1997 Electronic Visualization Laboratory - Univ Illinois at Chicago
 */

#include "CAVERNst_baseAvatar_c.h++"
#include "CAVERNplus_datapack_c.h++"
#include "_support_hash.h++"

int CAVERNst_baseAvatarManager_c::itsTrackerDataSize = 0;
int CAVERNst_baseAvatarManager_c::itsMetaDataSize = 0;

// Thread for avatar garbage collector.
nexus_thread_t g_cavernst_baseAvatarGarbageCollectThread;

// Flag to tell avatar garbage collector to quit.
int g_cavernst_baseAvatarGarbageCollectExitThread = 0;


CAVERNst_baseAvatar_c :: CAVERNst_baseAvatar_c(int avID,int trackerDataSize, int metaDataSize)
{
	avatarID = avID;
	trackerData = new char[trackerDataSize];
	metaData = new char[metaDataSize];
}

CAVERNst_baseAvatar_c :: ~CAVERNst_baseAvatar_c()
{
	if (trackerData) delete trackerData;
	if (metaData) delete metaData;
}

// Defined in cvrn_comm.c++
double cvrnGetTimeInSecs();

CAVERNst_baseAvatarManager_c :: ~CAVERNst_baseAvatarManager_c()
{
	/// Tell avatar thread to die.
	g_cavernst_baseAvatarGarbageCollectExitThread =  1;
	sendBye();

	/// Wait for the avatar garbage collect thread to die.
	while(g_cavernst_baseAvatarGarbageCollectExitThread != -1) sleep(1);

	if (avatarDB) delete avatarDB;
	if (hailingChannel) delete hailingChannel;
	if (trackerChannel) delete trackerChannel;
	if (ahcKey) delete ahcKey;
	if (atcKey) delete atcKey;
	if (hailingChannelLink) delete hailingChannelLink;
	if (trackerChannelLink) delete trackerChannelLink;
	if (selfAvatar) delete selfAvatar;
	if (ahcBuffer) delete ahcBuffer;
	if (atcBuffer) delete atcBuffer;
	if (tempBuffer) delete tempBuffer;
}


/* Main thread function for garbage collection.
Every 60 seconds this thread gathers a list of all the avatars and figures
out which ones have not sent any data in 120 seconds. Those are purged
and an exit callback is sent back to the user program.
*/
void *  cavernst_baseAvatarGarbageCollect(void *manager)
{
	CAVERNst_baseAvatarManager_c *theManager = (CAVERNst_baseAvatarManager_c *)manager;

	while(1) {

		/// Check for exit flag caused by a delete of the avatar manager.
		for (int ticker = 0; ticker < 60; ticker++) {
			sleep(1);
			if (g_cavernst_baseAvatarGarbageCollectExitThread){
				/// Let manager know that this thread will now terminate.
				g_cavernst_baseAvatarGarbageCollectExitThread = -1;
				return NULL;
			}
		}

		/// Every 60 seconds check for dead avatars.
		int numItems;

		/// Grab a list of avatars
		CAVERNst_baseAvatar_c **list;
		theManager->avatarDB->lock();
		list  = theManager->avatarDB->buildListOfEntries(&numItems);

		CAVERNst_baseAvatar_c *theAvatar;

		if (numItems) {
			for (int i = 0; i < numItems; i++) {
				theAvatar = list[i];

				/// If there are avatars that have not been refreshed in
				/// 2 minutes purge them.
				if ((cvrnGetTimeInSecs() - theAvatar->time) > 120) {
					theManager->avatarDB->remove(theAvatar->avatarID);
					cvrnMesg("CAVERNst_baseAvatarManager_c","Avatar %ld timed out",list[i]->avatarID);

					/// Call user callback
					if (theManager->avatarExitCB)
						theManager->avatarExitCB(theAvatar);


					delete theAvatar;
				}
			}
		}
		theManager->avatarDB->unlock();

		if (numItems) delete list;
	}

}


/* Initialize the avatar manager. It sets up keys for the Avatar hailing channel (AHC) and
the Avatar Tracker Channel (ATC). It also creates channels and links to remote IRBs if
one is specified.
 */
int CAVERNst_baseAvatarManager_c :: init(char* applicationName, CAVERN_irb_c *theIrb,
				       CAVERN_irbId_c *avatarIrbServer,
				       int trackerDataSize, int metaDataSize)
{

	if (trackerDataSize < 1) {
		cvrnMesg("CAVERNst_baseAvatarManager_c","Cannot set tracker data size to less than 1 byte.");
		exit(1);
	}

	if (metaDataSize < 1) {
		cvrnMesg("CAVERNst_baseAvatarManager_c","Cannot set meta data size to less than 1 byte.");
		exit(1);
	}

	// Allocate memory for packing and sending meta and trackerdata.
	ahcBufferSize =  1+sizeof(long) + metaDataSize;
	ahcBuffer = new char[ahcBufferSize];
	atcBufferSize =  1+sizeof(long) + trackerDataSize;
	atcBuffer = new char[ahcBufferSize];
	tempBuffer = new char[metaDataSize];

	requireSendAHCHello = 0;
	itsTrackerDataSize = trackerDataSize;
	itsMetaDataSize = metaDataSize;

	/// Generate AVATAR ID -----------------------------------------------------
	double theTime = cvrnGetTimeInSecs();

	CAVERN_irbId_c selfId;
	selfId = theIrb->getSelfIRBId();

	/// Avatar ID is composed of self IPaddress, port, and time of day in seconds
	char idStr[256];
	sprintf(idStr,"%s:%u:%lf", selfId.getAddress(), selfId.getPort(),
		theTime);

	/// Then this string is hashed into a unique 4 byte number.
	_support_hashWord_t hashVal = _support_Hash(idStr, strlen(idStr));
	long avatarID = hashVal;

	cvrnMesg("AVATAR INFO","YOUR AVATAR ID IS: %ld",avatarID);

	//-------------------------------------------------------------------------

	// Allocate memory for hash dictionary in which avatars are stored.
	avatarDB = new _support_hashDict<CAVERNst_baseAvatar_c *, int>;
	selfAvatar = new CAVERNst_baseAvatar_c(avatarID,trackerDataSize, metaDataSize);

	if (!selfAvatar) {
		cvrnMesg("CAVERNst_baseAvatarManager_c","Cannot allocate memory for self avatar.");
		exit(1);
	}


	// Save a pointer to personal IRB.
	itsIrb = theIrb;

	// IRB ID of remote IRB to connect to.
	if (avatarIrbServer) {
		itsAvatarIrbServer = *avatarIrbServer;
		selfServer = 0;
	}

	// If avatarIrbServer is NULL then this client is acting as the server
	// and so it makes no links to a remote IRB.
	else {
		selfServer = 1;
	}

	CAVERN_irb_c::status_t irbStatus;
	CAVERN_irbKeyId_c ahcKeyId, atcKeyId;

	// Define the local AHC and ATC keys.
	char appPath[256];

	// Use the application name to generate a path that
	// is unique for each application so that the keys
	// created for avatar data transport don't get mingled and avatars 
	// for different applications start popping up in the wrong apps.
	sprintf(appPath,"%s_avatar", applicationName);
	ahcKeyId.setPath(appPath);
	ahcKeyId.setName("AHC");
	ahcKey = itsIrb->define(&ahcKeyId,NULL,&irbStatus);
	ahcKey->trigger(ahcKeyCB, this);

	atcKeyId.setPath(appPath);
	atcKeyId.setName("ATC");
	atcKey = itsIrb->define(&atcKeyId,NULL,&irbStatus);
	atcKey->trigger(atcKeyCB, this);

	// If this client is not acting as a server open channels
	if (!selfServer) {

		hailingChannel = itsIrb->createChannel();
		trackerChannel = itsIrb->createChannel();

		CAVERN_irbChannel_c::status_t retStatus;
		hailingChannel->open(&itsAvatarIrbServer, NULL, CAVERN_irbChannel_c::RELIABLE,
				     &retStatus);
		if (retStatus != CAVERN_irbChannel_c::OK) exit(1);

		trackerChannel->open(&itsAvatarIrbServer, NULL, CAVERN_irbChannel_c::UNRELIABLE,
				     &retStatus);
		if (retStatus != CAVERN_irbChannel_c::OK) exit(1);

		// Make the links do NO initial synching.
		CAVERN_linkAttrib_c linkAttrib;
		linkAttrib.setInitialSynchType(CAVERN_linkAttrib_c::NONE);
		hailingChannelLink = hailingChannel->link(ahcKey, &ahcKeyId, &linkAttrib);
		if (!hailingChannelLink) {
			cvrnMesg("CAVERNst_baseAvatarManager_c","Unable to create a link for avatar hailing key to remote IRB key.");
			exit(1);
		}
		trackerChannelLink = trackerChannel->link(atcKey, &atcKeyId, &linkAttrib);
		if (!trackerChannelLink) {
			cvrnMesg("CAVERNst_baseAvatarManager_c","Unable to create a link for avatar tracker key to remote IRB key.");
			exit(1);
		}

	}

	/// Start the avatar garbage collection thread
	nexus_thread_create(&g_cavernst_baseAvatarGarbageCollectThread, NULL,
			    cavernst_baseAvatarGarbageCollect, this);

	return 1;
}

void CAVERNst_baseAvatarManager_c :: sendHello()
{
	putHelloToAHC();
}

void CAVERNst_baseAvatarManager_c :: sendTracker()
{
	putTrackerToATC();
}

void CAVERNst_baseAvatarManager_c :: sendBye()
{
	putByeToAHC();
	sleep(3);
}

/* Send a hello message that consists of:
   HELLO, AVATAR ID, META DATA
*/
int CAVERNst_baseAvatarManager_c :: putHelloToAHC()
{
	CAVERNplus_datapack_c datapack;


	datapack.setBuffer(ahcBuffer, ahcBufferSize);
	datapack.packChar(_AVATAR_HELLO_MSG);
	datapack.packLong(selfAvatar->avatarID);
	datapack.pack(selfAvatar->metaData,itsMetaDataSize);

	CAVERN_irbKey_c::status_t status;
	int size = datapack.getBufferFilledSize();
	ahcKey->put(ahcBuffer, &size, &status);

	return 1;
}

/*
  Send Bye message that consists of:
  BYE AVATAR ID
  */
int CAVERNst_baseAvatarManager_c :: putByeToAHC()
{
	CAVERNplus_datapack_c datapack;

	datapack.setBuffer(ahcBuffer, ahcBufferSize);
	datapack.packChar(_AVATAR_BYE_MSG);
	datapack.packLong(selfAvatar->avatarID);

	CAVERN_irbKey_c::status_t status;
	int size = datapack.getBufferFilledSize();
	ahcKey->put(ahcBuffer, &size, &status);

	return 1;
}

/*
  Send tracker data message which consists of:
  AVATAR_TRACKER_HEADER, avatarID, tracker data.
  */
int CAVERNst_baseAvatarManager_c :: putTrackerToATC()
{
	CAVERNplus_datapack_c datapack;

	datapack.setBuffer(atcBuffer, atcBufferSize);
	datapack.packChar(_AVATAR_TRACKER_MSG);
	datapack.packLong(selfAvatar->avatarID);
	datapack.pack(selfAvatar->trackerData, itsTrackerDataSize);

	CAVERN_irbKey_c::status_t status;
	int size;
	size = datapack.getBufferFilledSize();
	atcKey->put(atcBuffer, &size, &status);

	return 1;
}

// Callback for when avatar messages are received over the hailing channel.
void CAVERNst_baseAvatarManager_c :: ahcKeyCB(CAVERN_irbKey_c::CAVERN_irbKeyEvent_t,
						   CAVERN_irbKey_c *theKey,
						   void* manager)
{
	char *buf;
	int size;
	CAVERNst_baseAvatarManager_c *theAvatarManager;
	theAvatarManager = (CAVERNst_baseAvatarManager_c *) manager;

	CAVERN_irbKey_c::status_t status;
	theKey->getDirect(buf, &size, &status);
	CAVERNplus_datapack_c datapack;

	// Extract command and avatar id
	datapack.setBuffer(buf, size);
	char command = datapack.unpackChar();
	long avatarID = datapack.unpackLong();

	// Look for the avatar.
	CAVERNst_baseAvatar_c *theAvatar;
	theAvatarManager->avatarDB->lock();
	theAvatarManager->avatarDB->find(avatarID, theAvatar);
	theAvatarManager->avatarDB->unlock();

	if (command == _AVATAR_HELLO_MSG) {

		// If avatar does not exist yet make one.
		if (!theAvatar) {

			theAvatar = new CAVERNst_baseAvatar_c(avatarID, itsTrackerDataSize,
							    itsMetaDataSize);
			if (!theAvatar) {
				cvrnMesg("CAVERNst_baseAvatar_c","Unable to allocate new entry for avatar.");
				theKey->releaseDirect(&status);
				return;
			}

			/// Refresh the timer so that this avatar does not get garbage collected.
			theAvatar->time = cvrnGetTimeInSecs();

			// Store avatar in avatar database
			theAvatarManager->avatarDB->lock();
			theAvatarManager->avatarDB->enter(avatarID, theAvatar);
			theAvatarManager->avatarDB->unlock();
			datapack.unpack(theAvatar->metaData, itsMetaDataSize);
			theKey->releaseDirect(&status);

			cvrnMesg("CAVERNst_baseAvatar_c","NEW AVATAR IN TOWN %d",avatarID);

			// Call the user's callback if any
			if (theAvatarManager->newAvatarCB)
				theAvatarManager->newAvatarCB(theAvatar);

			// Set a flag so that a hello message can be sent out.
			if (command == _AVATAR_HELLO_MSG)
				theAvatarManager->requireSendAHCHello = 1;

		} else {
			/// Refresh the timer so that this avatar does not get garbage collected.
			theAvatar->time = cvrnGetTimeInSecs();

			// Grab the meta data.
			datapack.unpack(theAvatarManager->tempBuffer, itsMetaDataSize);
			theKey->releaseDirect(&status);

			// Is it different from previous value?
			if (bcmp(theAvatarManager->tempBuffer, theAvatar->metaData, itsMetaDataSize)) {

				// Replace with new one.
				memcpy(theAvatar->metaData, theAvatarManager->tempBuffer, itsMetaDataSize);

				// Call user callback.
				if (theAvatarManager->avatarMetaChangeCB)
					theAvatarManager->avatarMetaChangeCB(theAvatar);
			}
		}
	} else if (command == _AVATAR_BYE_MSG) {

		theKey->releaseDirect(&status);

		if (!theAvatar) return;

		theAvatarManager->avatarDB->lock();
		theAvatarManager->avatarDB->remove(avatarID);
		theAvatarManager->avatarDB->unlock();

		if (theAvatarManager->avatarExitCB) theAvatarManager->avatarExitCB(theAvatar);
		delete theAvatar;

		return;
	}

}

void CAVERNst_baseAvatarManager_c :: setMetaData(char *meta)
{
	memcpy(selfAvatar->metaData,meta, itsMetaDataSize);
}

void CAVERNst_baseAvatarManager_c :: setTrackerData(char *tracker)
{
	memcpy(selfAvatar->trackerData,tracker, itsTrackerDataSize);
}


// Callback for when avatar tracker data comes in.
void CAVERNst_baseAvatarManager_c :: atcKeyCB(CAVERN_irbKey_c::CAVERN_irbKeyEvent_t,
						   CAVERN_irbKey_c *theKey,
						   void* manager)
{
	char *buf;
	int size;
	CAVERNst_baseAvatarManager_c *theAvatarManager;
	theAvatarManager = (CAVERNst_baseAvatarManager_c *) manager;

	// This is an opportunity to send out a hello message based on
	// whether someone had already sent us a hello message.
	if (theAvatarManager->requireSendAHCHello) {
		theAvatarManager->putHelloToAHC();
		theAvatarManager->requireSendAHCHello = 0;
	}

	// Grab the command and avatar id.
	CAVERN_irbKey_c::status_t status;
	theKey->getDirect(buf, &size, &status);
	CAVERNplus_datapack_c datapack;

	datapack.setBuffer(buf, size);
	char command = datapack.unpackChar();
	long avatarID = datapack.unpackLong();

	// Look for the avatar.
	CAVERNst_baseAvatar_c *theAvatar;
	theAvatarManager->avatarDB->lock();
	theAvatarManager->avatarDB->find(avatarID, theAvatar);
	theAvatarManager->avatarDB->unlock();

	// If it's not found then ignore the packet.
	if (!theAvatar) {
		theKey->releaseDirect (&status);
		return;
	}

	// If it is found then retrieve the avatar tracker data.
	if (command == _AVATAR_TRACKER_MSG) {
		theAvatar->time = cvrnGetTimeInSecs();

		datapack.unpack(theAvatar->trackerData, itsTrackerDataSize);

		// Fire the user's callback
		if (theAvatarManager->avatarTrackerCB)
			theAvatarManager->avatarTrackerCB(theAvatar);

	} 
	theKey->releaseDirect(&status);

}
