/* Author: Jason Leigh
 * (C) 1997 Electronic Visualization Laboratory - Univ Illinois at Chicago
 */
#include "CAVERN.h++"


/*
2 channels, avatar hailing channel (AHC) and avatar tracker channel (ATC).

Messages over AHC:
[HELLO, avatarID, metaData]
[BYE, avatarID]

Messages over ATC:
[TRACKERDATA, avatarID, trackerData]

*/


#define _AVATAR_HELLO_MSG 'h'
#define _AVATAR_BYE_MSG 'b'
#define _AVATAR_TRACKER_MSG 't'

// An auxiliary message sent over the hailing channel that does not involve
// the creation of avatars.
#define _AVATAR_AUX_MSG 'a'

/** CAVERN support template base avatar class.
Objects of this type are passed in as parameters of user-specified callbacks
to the avatar manager when certain events occur.
You can then look directly into the fields to find the info you need.
  */
class CAVERNst_baseAvatar_c {
public:
	/// Avatar meta data.
	char *metaData;

	/// Avatar tracker data.
	char *trackerData;

	/// This avatar's ID
	long avatarID;

	/// Time of last change in seconds.
	double time;

	CAVERNst_baseAvatar_c(int avID, int trackerDataSize, int metaDataSize);
	~CAVERNst_baseAvatar_c();
};

/** CAVERN support template base avatar manager class.
This manages avatars that connect and disconnect from your client.
Begin by creating a single manager and then initializing it with init().
Then set the meta data you wish to send and call sendHello().
Everytime you wish to send a tracker packet, package the tracker data
you want to send, call setTrackerData() and then call sendTracker().
When you want to tell everyone you are leaving call: sendBye()
or simply delete your avatar manager object.

Note: nothing is implied in the format of the tracker or meta data.
This is done intentionally so that you can bundle whatever you want in it.
Then you can build higher level avatar classes on top of this or extend this
avatar class by adding things that are specific to say OpenGL or Performer.

For example, tracker data could be an array of floats containing an avatar's
body position, orientation etc.
Meta data could consist of information about an avatar's color, size,
filenames for 3D models, IRB addresses and keys to find 3D models, etc.

*/
class CAVERNst_baseAvatarManager_c {
public:

	/** Initialize Avatar Manager. Do this before doing any other avatar manager calls.
	    @param applicationName a string uniquely identifying an application group. This
	    is to prevent avatars from different applications from running into each other.
	    @param theIRB the personal IRB from CAVERNInit()
	    @param avatarIrbServer the IRB ID of an IRB to which you wish to designate as
	    a server to handle avatar information distribution. If you set this as NULL
	    then you are assuming that responsibility yourself and you can ask other
	    clients to connect to you.
	    @param trackerDataSize the expected size of the data that will carry all the 
	    positional, and orientational information about the avatar all packed into
	    1  array of characters.
	    @param metaDataSize the expected size of the data that will carry all the
	    other avatar information like the color of an avatar, or the filename for
	    each part of the avatar (head, hand etc), or the IRB address of where
	    to retrieve the different parts of the avatar.
	    */
	int init(char* applicationName, CAVERN_irb_c *theIrb,
		 CAVERN_irbId_c *avatarIrbServer,
		 int trackerDataSize,
		 int metaDataSize);

	/** Delete the avatar manager and also tell other clients that you
	have left.
	*/
	~CAVERNst_baseAvatarManager_c();

	/// When a new avatar arrives call your callback.
	void setNewAvatarCallback(void (*cb)(CAVERNst_baseAvatar_c *)) {newAvatarCB = cb;}

	/// When avatar meta data has changed call your callback.
	void setAvatarMetaChangeCallback(void (*cb)(CAVERNst_baseAvatar_c *)) {avatarMetaChangeCB = cb;}

	/// If an avatar has left call your callback.
	void setAvatarExitCallback(void (*cb)(CAVERNst_baseAvatar_c *)) {avatarExitCB = cb;}

	/// If a new packet of tracker data arrives call your callback.
	void setAvatarTrackerCallback(void (*cb)(CAVERNst_baseAvatar_c *)) {avatarTrackerCB = cb;}

	/// Set the meta data that you wish to send out. Use sendHello() to dispatch the data.
	void setMetaData(char *meta);

	/// Set the tracker data that you wish to send out. Use sendTracker() to dispatch the data.
	void setTrackerData(char *tracker);

	/// Sending the hello also sends the meta data.
	void sendHello();

	/// Sending the tracker only sends the tracker data.
	void sendTracker();

	/// Send bye message to all clients.
	void sendBye();

	_support_hashDict<CAVERNst_baseAvatar_c *, int> *avatarDB;

	void (*avatarExitCB) (CAVERNst_baseAvatar_c*);

private:
	int requireSendAHCHello;

	static int itsTrackerDataSize, itsMetaDataSize;
	char *ahcBuffer, *tempBuffer, *atcBuffer;
	int ahcBufferSize, atcBufferSize;


	CAVERN_irb_c *itsIrb;
	CAVERN_irbId_c itsAvatarIrbServer;

	// Determines if you are going to be the main avatar server.
	int selfServer;

	CAVERN_irbChannel_c *hailingChannel;
	CAVERN_irbChannel_c *trackerChannel;
	CAVERN_irbLink_c *hailingChannelLink;
	CAVERN_irbLink_c *trackerChannelLink;
	CAVERN_irbKey_c *ahcKey,  *atcKey;

	CAVERNst_baseAvatar_c *selfAvatar;

	// Callback fired when a message comes in over the AHC.
	static void ahcKeyCB(CAVERN_irbKey_c::CAVERN_irbKeyEvent_t,
		      CAVERN_irbKey_c *theKey,
		      void*);

	// Callback fired when a message comes in over the ATC.
	static void atcKeyCB(CAVERN_irbKey_c::CAVERN_irbKeyEvent_t,
		      CAVERN_irbKey_c *theKey,
		      void*);



	void (*newAvatarCB) (CAVERNst_baseAvatar_c*);
	void (*avatarMetaChangeCB) (CAVERNst_baseAvatar_c*);
	void (*avatarTrackerCB) (CAVERNst_baseAvatar_c*);

	int putHelloToAHC();
	int putTrackerToATC();
	int putByeToAHC();

};
